# :coding: utf-8

import pytest

import champollion.parser.js_class


@pytest.mark.parametrize(
    ("content", "expected"),
    [
        (
            (
                "/**\n"
                " * Simple class\n"
                " */\n"
                "class SimpleClass {}\n"
            ),
            {
                "test.module.SimpleClass": {
                    "id": "test.module.SimpleClass",
                    "module_id": "test.module",
                    "exported": False,
                    "default": False,
                    "name": "SimpleClass",
                    "parent": None,
                    "line_number": 4,
                    "description": "Simple class",
                    "method": {},
                    "attribute": {}
                }
            }
        ),
        (
            (
                "/**\n"
                " * Simple class with attributes\n"
                " */\n"
                "class SimpleClass {\n"
                "    /**\n"
                "     * A static attribute.\n"
                "     */\n"
                "    static attribute1 = 42;\n"
                "\n"
                "    /**\n"
                "     * An object attribute.\n"
                "     */\n"
                "    attribute2 = {\n"
                "        key1: 'value1',\n"
                "        key2: 'value2',\n"
                "    };\n"
                "\n"
                "    /**\n"
                "     * A list attribute.\n"
                "     */\n"
                "    attribute3 = [\n"
                "        'value1',\n"
                "        'value2',\n"
                "    ];\n"
                "}\n"
            ),
            {
                "test.module.SimpleClass": {
                    "id": "test.module.SimpleClass",
                    "module_id": "test.module",
                    "exported": False,
                    "default": False,
                    "name": "SimpleClass",
                    "parent": None,
                    "line_number": 4,
                    "description": "Simple class with attributes",
                    "method": {},
                    "attribute": {
                        "test.module.SimpleClass.attribute1": {
                            "id": "test.module.SimpleClass.attribute1",
                            "class_id": "test.module.SimpleClass",
                            "module_id": "test.module",
                            "name": "attribute1",
                            "prefix": "static",
                            "value": "42",
                            "line_number": 8,
                            "description": "A static attribute."
                        },
                        "test.module.SimpleClass.attribute2": {
                            "id": "test.module.SimpleClass.attribute2",
                            "class_id": "test.module.SimpleClass",
                            "module_id": "test.module",
                            "name": "attribute2",
                            "prefix": None,
                            "value": "{ key1: 'value1', key2: 'value2', }",
                            "line_number": 13,
                            "description": "An object attribute."
                        },
                        "test.module.SimpleClass.attribute3": {
                            "id": "test.module.SimpleClass.attribute3",
                            "class_id": "test.module.SimpleClass",
                            "module_id": "test.module",
                            "name": "attribute3",
                            "prefix": None,
                            "value": "[ 'value1', 'value2', ]",
                            "line_number": 21,
                            "description": "A list attribute."
                        }
                    }
                }
            }
        ),
        (
            (
                "/**\n"
                " * Simple class with getter and setter\n"
                " */\n"
                "class SimpleClass {\n"
                "    /**\n"
                "     * Get the name.\n"
                "     *\n"
                "     * .. warning::\n"
                "     *\n"
                "     *     The name is awesome\n"
                "     */\n"
                "    get name() {\n"
                "        return this.name;\n"
                "    }\n"
                "\n"
                "    /**\n"
                "     * Set the name.\n"
                "     *\n"
                "     * .. warning::\n"
                "     *\n"
                "     *     Keep the name awesome\n"
                "     */\n"
                "    set name(value) {\n"
                "        this.name = value;\n"
                "    }\n"
                "}\n"
            ),
            {
                "test.module.SimpleClass": {
                    "id": "test.module.SimpleClass",
                    "module_id": "test.module",
                    "exported": False,
                    "default": False,
                    "name": "SimpleClass",
                    "parent": None,
                    "line_number": 4,
                    "description": "Simple class with getter and setter",
                    "method": {
                        "test.module.SimpleClass.name.get": {
                            "id": "test.module.SimpleClass.name.get",
                            "class_id": "test.module.SimpleClass",
                            "module_id": "test.module",
                            "name": "name",
                            "prefix": "get",
                            "arguments": [],
                            "line_number": 12,
                            "description": (
                                "Get the name.\n"
                                "\n"
                                ".. warning::\n"
                                "\n"
                                "    The name is awesome"
                            )
                        },
                        "test.module.SimpleClass.name.set": {
                            "id": "test.module.SimpleClass.name.set",
                            "class_id": "test.module.SimpleClass",
                            "module_id": "test.module",
                            "name": "name",
                            "prefix": "set",
                            "arguments": ["value"],
                            "line_number": 23,
                            "description": (
                                "Set the name.\n"
                                "\n"
                                ".. warning::\n"
                                "\n"
                                "    Keep the name awesome"
                            )
                        }
                    },
                    "attribute": {}
                }
            }
        ),
        (
            (
                "/**\n"
                " * Simple class with constructor\n"
                " */\n"
                "class SimpleClass {\n"
                "    constructor() {\n"
                "    }\n"
                "}\n"
            ),
            {
                "test.module.SimpleClass": {
                    "id": "test.module.SimpleClass",
                    "module_id": "test.module",
                    "exported": False,
                    "default": False,
                    "name": "SimpleClass",
                    "parent": None,
                    "line_number": 4,
                    "description": "Simple class with constructor",
                    "method": {
                        "test.module.SimpleClass.constructor": {
                            "id": "test.module.SimpleClass.constructor",
                            "class_id": "test.module.SimpleClass",
                            "module_id": "test.module",
                            "name": "constructor",
                            "prefix": None,
                            "arguments": [],
                            "line_number": 5,
                            "description": None
                        }
                    },
                    "attribute": {}
                }
            }
        ),
        (
            (
                "/** Simple class expression */\n"
                "export const CustomWelcome = class Welcome extends Base{\n"
                "    static expression= 'Hello World';\n"
                "\n"
                "    /** Say Hi to the world */\n"
                "    greeting() {\n"
                "        return this.expression;\n"
                "    }\n"
                "};\n"
                "\n"
            ),
            {
                "test.module.CustomWelcome": {
                    "id": "test.module.CustomWelcome",
                    "module_id": "test.module",
                    "exported": True,
                    "default": False,
                    "name": "CustomWelcome",
                    "parent": "Base",
                    "line_number": 2,
                    "description": "Simple class expression",
                    "method": {
                        "test.module.CustomWelcome.greeting": {
                            "id": "test.module.CustomWelcome.greeting",
                            "class_id": "test.module.CustomWelcome",
                            "module_id": "test.module",
                            "name": "greeting",
                            "prefix": None,
                            "arguments": [],
                            "line_number": 6,
                            "description": "Say Hi to the world"
                        }
                    },
                    "attribute": {
                        "test.module.CustomWelcome.expression": {
                            "id": "test.module.CustomWelcome.expression",
                            "class_id": "test.module.CustomWelcome",
                            "module_id": "test.module",
                            "name": "expression",
                            "prefix": "static",
                            "value": "'Hello World'",
                            "line_number": 3,
                            "description": None
                        }
                    }
                }
            }
        ),
        (
            (
                (
                    "/** Simple class */\n"
                    "class Welcome {\n"
                    "    /** Say Hi to someone */\n"
                    "    greeting = who =>\n"
                    "        `Hello ${who}!`;\n"
                    "};\n"
                    "\n"
                ),
                {
                    "test.module.Welcome": {
                        "id": "test.module.Welcome",
                        "module_id": "test.module",
                        "exported": False,
                        "default": False,
                        "name": "Welcome",
                        "parent": None,
                        "line_number": 2,
                        "description": "Simple class",
                        "method": {
                            "test.module.Welcome.greeting": {
                                "id": "test.module.Welcome.greeting",
                                "class_id": "test.module.Welcome",
                                "module_id": "test.module",
                                "name": "greeting",
                                "prefix": None,
                                "arguments": ["who"],
                                "line_number": 4,
                                "description": "Say Hi to someone"
                            }
                        },
                        "attribute": {
                            "test.module.Welcome.greeting": {
                                "id": "test.module.Welcome.greeting",
                                "class_id": "test.module.Welcome",
                                "module_id": "test.module",
                                "name": "greeting",
                                "prefix": None,
                                "value": "who =>`Hello ${who}!`",
                                "line_number": 4,
                                "description": "Say Hi to someone"
                            }
                        }
                    },
                }
            )
        )
    ],
    ids=[
        "valid class",
        "valid class with attributes",
        "valid class with getter and setter methods",
        "valid class with constructor",
        "valid exported class expression with mother class",
        "valid class with arrow-type method and a single argument",
    ]
)
def test_get_class_environment(content, expected):
    """Return class environment from content."""
    assert champollion.parser.js_class.fetch_environment(
        content, "test.module"
    ) == expected


@pytest.mark.parametrize(
    ("content", "expected"),
    [
        (
            "class AwesomeClass",
            None
        ),
        (
            "const test= 'class AwesomeClass {}';",
            None
        ),
        (
            "class AwesomeClass {}",
            {
                "export": None,
                "default": None,
                "class_name": "AwesomeClass",
                "data_name": None,
                "mother_class": None,
                "start_regex": ""
            }
        ),
        (
            "class Awesome_Class extends module.Mother-Class{}",
            {
                "export": None,
                "default": None,
                "class_name": "Awesome_Class",
                "data_name": None,
                "mother_class": "module.Mother-Class",
                "start_regex": ""
            }
        ),
        (
            "export default class AwesomeClass {}",
            {
                "export": "export ",
                "default": "default ",
                "class_name": "AwesomeClass",
                "data_name": None,
                "mother_class": None,
                "start_regex": ""
            }
        ),
        (
            "export const MyClass1= class AwesomeClass {}",
            {
                "export": "export ",
                "default": None,
                "class_name": None,
                "data_name": "MyClass1",
                "mother_class": None,
                "start_regex": ""
            }
        ),
        (
            "let MyClass1= class AwesomeClass extends Test2 {}",
            {
                "export": None,
                "default": None,
                "class_name": None,
                "data_name": "MyClass1",
                "mother_class": "Test2",
                "start_regex": ""
            }
        ),
    ],
    ids=[
        "invalid class",
        "invalid class string",
        "valid class",
        "valid class with inheritance",
        "valid class exported by default",
        "valid class expression exported",
        "another valid class expression",
    ]
)
def test_class_pattern(content, expected):
    """Match a class."""
    match = champollion.parser.js_class._CLASS_PATTERN.search(content)
    if expected is None:
        assert match is None
    else:
        assert match.groupdict() == expected


@pytest.mark.parametrize(
    ("content", "expected"),
    [
        (
            "valid-Method (arg1) {}",
            {
                "arguments": "arg1",
                "method_name": "valid-Method",
                "prefix": None,
                "start_regex": ""
            }
        ),
        (
            "validMethod (\n"
            "    arg1, arg2, arg3, arg4, arg5,\n"
            "    arg6, arg7, arg8, arg9, arg10\n"
            ") {}",
            {
                "arguments": (
                    "arg1, arg2, arg3, arg4, arg5,\n"
                    "    arg6, arg7, arg8, arg9, arg10"
                ),
                "method_name": "validMethod",
                "prefix": None,
                "start_regex": ""
            }
        ),
        (
            "static valid_method() {}",
            {
                "arguments": "",
                "method_name": "valid_method",
                "prefix": "static ",
                "start_regex": ""
            }
        ),
        (
            "get valid_method2(){}",
            {
                "arguments": "",
                "method_name": "valid_method2",
                "prefix": "get ",
                "start_regex": ""
            }
        ),
        (
            (
                "set validMethod( \n"
                "    arg1, arg2, arg3, arg4, arg5,\n"
                "){\n"
                "    console.log('test');\n"
                "}\n"
            ),
            {
                "arguments": "arg1, arg2, arg3, arg4, arg5,",
                "method_name": "validMethod",
                "prefix": "set ",
                "start_regex": ""
            }
        ),
        (
            "function invalidMethod() {}",
            None
        ),
        (
            "invalidMethod()",
            None
        ),
        (
            "invalidMethod {}",
            None
        ),
    ],
    ids=[
        "valid method",
        "valid method with many arguments",
        "valid static method",
        "valid getter method",
        "valid setter method",
        "invalid method with 'function' statement",
        "invalid method without nested element",
        "invalid method without argument",
    ]
)
def test_class_method_pattern(content, expected):
    """Match a class method."""
    match = champollion.parser.js_class._CLASS_METHOD_PATTERN.search(
        content
    )
    if expected is None:
        assert match is None
    else:
        assert match.groupdict() == expected


@pytest.mark.parametrize(
    ("content", "expected"),
    [
        (
            "arrow_type_method = (arg1) => {};",
            {
                "arguments": "arg1",
                "single_argument": None,
                "method_name": "arrow_type_method",
                "prefix": None,
                "start_regex": ""
            }
        ),
        (
            "arrow_type_method2 = arg1 => {};",
            {
                "arguments": None,
                "single_argument": "arg1",
                "method_name": "arrow_type_method2",
                "prefix": None,
                "start_regex": ""
            }
        ),
        (
            "static arrow_type_method = (arg1) => {};",
            {
                "arguments": "arg1",
                "single_argument": None,
                "method_name": "arrow_type_method",
                "prefix": "static ",
                "start_regex": ""
            }
        ),
        (
            "arrow_type_method3 = (arg1, arg2) => {};",
            {
                "arguments": "arg1, arg2",
                "single_argument": None,
                "method_name": "arrow_type_method3",
                "prefix": None,
                "start_regex": ""
            }
        ),
        (
            (
                "arrow_type_method3 = (\n"
                "    arg1, arg2, arg3, arg4, arg5, agr6,\n"
                "    arg7\n"
                ") => {\n"
                "    console.log('youpi');\n"
                "};\n"
            ),
            {
                "arguments": "arg1, arg2, arg3, arg4, arg5, agr6,\n    arg7",
                "single_argument": None,
                "method_name": "arrow_type_method3",
                "prefix": None,
                "start_regex": ""
            }
        ),
        (
            "const arrow_type_method = (arg1) => {}",
            None
        ),
        (
            "const test = 'arrow_type_method = (arg1) => {}'",
            None
        ),
        (
            "(arg1) => {}",
            None
        ),
        (
            "const arrow_type_method = arg1, arg2 => {};",
            None
        ),
    ],
    ids=[
        "valid method with one argument and brackets",
        "valid method with one argument and no brackets",
        "valid static method",
        "valid method with two arguments",
        "valid method with multiple arguments",
        "invalid arrow-type method with type",
        "invalid arrow-type method string",
        "invalid unassigned arrow-type method",
        "invalid arrow-type method with multiple argument and no brackets",
    ]
)
def test_class_method_arrow_pattern(content, expected):
    """Match a class arrow-type method."""
    match = champollion.parser.js_class._CLASS_METHOD_ARROW_PATTERN.search(
        content
    )
    if expected is None:
        assert match is None
    else:
        assert match.groupdict() == expected


@pytest.mark.parametrize(
    ("content", "expected"),
    [
        (
            "attribute_test1 = 42;",
            {
                "name": "attribute_test1",
                "value": "42;",
                "prefix": None,
                "start_regex": ""
            }
        ),
        (
            (
                "static attribute_test2 = {\n"
                "    key: 'value',\n"
                "};"
            ),
            {
                "name": "attribute_test2",
                "value": (
                    "{\n"
                    "    key: 'value',\n"
                    "};"
                ),
                "prefix": "static ",
                "start_regex": ""
            }
        ),
        (
            (
                "attribute_test3 = [\n"
                "    1, 2, 3, 4, 5, 6, 7, 8, 9, 10,\n"
                "];"
            ),
            {
                "name": "attribute_test3",
                "value": (
                    "[\n"
                    "    1, 2, 3, 4, 5, 6, 7, 8, 9, 10,\n"
                    "];"
                ),
                "prefix": None,
                "start_regex": ""
            }
        ),
        (
            "const test = 'static attribute = 42';",
            None
        ),
        (
            "attribute_test1 = 42",
            None
        ),

    ],
    ids=[
        "valid attribute",
        "valid static object attribute",
        "valid static object attribute",
        "invalid attribute string",
        "invalid attribute with no semi-colons",
    ]
)
def test_class_attribute_pattern(content, expected):
    """Match a class attribute."""
    match = champollion.parser.js_class._CLASS_ATTRIBUTE_PATTERN.search(
        content
    )
    if expected is None:
        assert match is None
    else:
        assert match.groupdict() == expected



# :coding: utf-8

import pytest
import os
import re
import unicodedata

from sphinx.cmdline import main as sphinx_main
from sphinx.util.osutil import cd


def _sanitise_value(value):
    """Return *value* suitable for comparison using python 2 and python 3.
    """
    value = value.decode("UTF-8")
    value = unicodedata.normalize("NFKD", value)
    value = value.encode("ascii", "ignore").decode("UTF-8")
    value = re.sub(
        r"[^\w*._\-\\/:% \"()\[\]{}\n=,]", "", value
    )
    return value


@pytest.fixture()
def doc_folder_with_code(doc_folder):
    """Return Doc folder with Javascript example source code.
    """
    js_source = os.path.join(doc_folder, "example")

    with open(os.path.join(js_source, "index.js"), "w") as f:
        f.write(
            "/**\n"
            " * A function\n"
            " *\n"
            " * .. note::\n"
            " *\n"
            " *     A note.\n"
            " */\n"
            "export default function doSomething1(arg1, arg2 = null) {\n"
            "    console.log('test1')\n"
            "}\n"
            "\n"
            "/**\n"
            " * Another function\n"
            " *\n"
            " * A citation::\n"
            " *\n"
            " *     A citation\n"
            " */\n"
            "const doSomething2 = (arg) => {\n"
            "    console.log('test2')\n"
            "};\n"
            "\n"
            "export const doSomething3 = () => {};\n"
            "\n"
            "export default function() {}\n"
            "\n"
            "/** generator function */\n"
            "const yieldSomethingAliased = function* yieldSomething(arg) {}\n"
        )

    return doc_folder


def test_directive_autofunction(doc_folder_with_code):
    """Generate documentation from functions.
    """
    index_file = os.path.join(doc_folder_with_code, "index.rst")
    with open(index_file, "w") as f:
        f.write(
            ".. js:autofunction:: example.doSomething1\n"
            "\n"
            ".. js:autofunction:: example.doSomething2\n"
            "\n"
            ".. js:autofunction:: example.doSomething3\n"
            "\n"
            ".. js:autofunction:: example.__ANONYMOUS_FUNCTION__\n"
            "\n"
            ".. js:autofunction:: example.yieldSomethingAliased\n"
        )

    with cd(doc_folder_with_code):
        sphinx_main(["-c", ".", "-b", "text", "-E", ".", "_build"])

    with open(
        os.path.join(doc_folder_with_code, "_build", "index.txt"), "rb"
    ) as f:
        content = _sanitise_value(f.read())

        assert content == (
            "example.doSomething1(arg1, arg2 = null)\n"
            "\n"
            "   \"import doSomething1 from \"example\"\"\n"
            "\n"
            "   A function\n"
            "\n"
            "   Note: A note.\n"
            "\n"
            "example.doSomething2(arg)\n"
            "\n"
            "   Another function\n"
            "\n"
            "   A citation:\n"
            "\n"
            "      A citation\n"
            "\n"
            "example.doSomething3()\n"
            "\n"
            "   \"import {doSomething3} from \"example\"\"\n"
            "\n"
            "example.__ANONYMOUS_FUNCTION__()\n"
            "\n"
            "function* example.yieldSomethingAliased(arg)\n"
            "\n"
            "   generator function\n"
        )


def test_directive_autofunction_with_alias(doc_folder_with_code):
    """Generate documentation from functions with alias.
    """
    index_file = os.path.join(doc_folder_with_code, "index.rst")
    with open(index_file, "w") as f:
        f.write(
            ".. js:autofunction:: example.doSomething1\n"
            "    :alias: aliased_doSomething1\n"
            "\n"
            ".. js:autofunction:: example.doSomething2\n"
            "    :alias: aliased_doSomething2\n"
            "\n"
            ".. js:autofunction:: example.doSomething3\n"
            "    :alias: aliased_doSomething3\n"
            "\n"
            ".. js:autofunction:: example.__ANONYMOUS_FUNCTION__\n"
            "    :alias: alias_anonymous\n"
            "\n"
            ".. js:autofunction:: example.yieldSomethingAliased\n"
            "    :alias: aliased_generate\n"
        )

    with cd(doc_folder_with_code):
        sphinx_main(["-c", ".", "-b", "text", "-E", ".", "_build"])

    with open(
        os.path.join(doc_folder_with_code, "_build", "index.txt"), "rb"
    ) as f:
        content = _sanitise_value(f.read())

        assert content == (
            "example.aliased_doSomething1(arg1, arg2 = null)\n"
            "\n"
            "   \"import aliased_doSomething1 from \"example\"\"\n"
            "\n"
            "   A function\n"
            "\n"
            "   Note: A note.\n"
            "\n"
            "example.aliased_doSomething2(arg)\n"
            "\n"
            "   Another function\n"
            "\n"
            "   A citation:\n"
            "\n"
            "      A citation\n"
            "\n"
            "example.aliased_doSomething3()\n"
            "\n"
            "   \"import {aliased_doSomething3} from \"example\"\"\n"
            "\n"
            "example.alias_anonymous()\n"
            "\n"
            "function* example.aliased_generate(arg)\n"
            "\n"
            "   generator function\n"
        )


def test_directive_autofunction_with_module_alias(doc_folder_with_code):
    """Generate documentation from functions with module alias.
    """
    index_file = os.path.join(doc_folder_with_code, "index.rst")
    with open(index_file, "w") as f:
        f.write(
            ".. js:autofunction:: example.doSomething1\n"
            "    :module-alias: alias_module\n"
            "\n"
            ".. js:autofunction:: example.doSomething2\n"
            "    :module-alias: alias_module\n"
            "\n"
            ".. js:autofunction:: example.doSomething3\n"
            "    :module-alias: alias_module\n"
            "\n"
            ".. js:autofunction:: example.__ANONYMOUS_FUNCTION__\n"
            "    :module-alias: alias_module\n"
            "\n"
            ".. js:autofunction:: example.yieldSomethingAliased\n"
            "    :module-alias: alias_module\n"
        )

    with cd(doc_folder_with_code):
        sphinx_main(["-c", ".", "-b", "text", "-E", ".", "_build"])

    with open(
        os.path.join(doc_folder_with_code, "_build", "index.txt"), "rb"
    ) as f:
        content = _sanitise_value(f.read())

        assert content == (
            "alias_module.doSomething1(arg1, arg2 = null)\n"
            "\n"
            "   \"import doSomething1 from \"example\"\"\n"
            "\n"
            "   A function\n"
            "\n"
            "   Note: A note.\n"
            "\n"
            "alias_module.doSomething2(arg)\n"
            "\n"
            "   Another function\n"
            "\n"
            "   A citation:\n"
            "\n"
            "      A citation\n"
            "\n"
            "alias_module.doSomething3()\n"
            "\n"
            "   \"import {doSomething3} from \"example\"\"\n"
            "\n"
            "alias_module.__ANONYMOUS_FUNCTION__()\n"
            "\n"
            "function* alias_module.yieldSomethingAliased(arg)\n"
            "\n"
            "   generator function\n"
        )


def test_directive_autofunction_with_module_path_alias(doc_folder_with_code):
    """Generate documentation from functions with module path alias.
    """
    index_file = os.path.join(doc_folder_with_code, "index.rst")
    with open(index_file, "w") as f:
        f.write(
            ".. js:autofunction:: example.doSomething1\n"
            "    :module-path-alias: test/alias/module\n"
            "\n"
            ".. js:autofunction:: example.doSomething2\n"
            "    :module-path-alias: test/alias/module\n"
            "\n"
            ".. js:autofunction:: example.doSomething3\n"
            "    :module-path-alias: test/alias/module\n"
            "\n"
            ".. js:autofunction:: example.__ANONYMOUS_FUNCTION__\n"
            "    :module-path-alias: test/alias/module\n"
            "\n"
            ".. js:autofunction:: example.yieldSomethingAliased\n"
            "    :module-path-alias: test/alias/module\n"
        )

    with cd(doc_folder_with_code):
        sphinx_main(["-c", ".", "-b", "text", "-E", ".", "_build"])

    with open(
        os.path.join(doc_folder_with_code, "_build", "index.txt"), "rb"
    ) as f:
        content = _sanitise_value(f.read())

        assert content == (
            "example.doSomething1(arg1, arg2 = null)\n"
            "\n"
            "   \"import doSomething1 from \"test/alias/module\"\"\n"
            "\n"
            "   A function\n"
            "\n"
            "   Note: A note.\n"
            "\n"
            "example.doSomething2(arg)\n"
            "\n"
            "   Another function\n"
            "\n"
            "   A citation:\n"
            "\n"
            "      A citation\n"
            "\n"
            "example.doSomething3()\n"
            "\n"
            "   \"import {doSomething3} from \"test/alias/module\"\"\n"
            "\n"
            "example.__ANONYMOUS_FUNCTION__()\n"
            "\n"
            "function* example.yieldSomethingAliased(arg)\n"
            "\n"
            "   generator function\n"
        )


def test_directive_autofunction_with_partial_import_forced(
    doc_folder_with_code
):
    """Generate documentation from functions with partial import forced.
    """
    index_file = os.path.join(doc_folder_with_code, "index.rst")
    with open(index_file, "w") as f:
        f.write(
            ".. js:autofunction:: example.doSomething1\n"
            "    :force-partial-import:\n"
            "\n"
            ".. js:autofunction:: example.doSomething2\n"
            "    :force-partial-import:\n"
            "\n"
            ".. js:autofunction:: example.doSomething3\n"
            "    :force-partial-import:\n"
            "\n"
            ".. js:autofunction:: example.__ANONYMOUS_FUNCTION__\n"
            "    :force-partial-import:\n"
            "\n"
            ".. js:autofunction:: example.yieldSomethingAliased\n"
            "    :force-partial-import:\n"
        )

    with cd(doc_folder_with_code):
        sphinx_main(["-c", ".", "-b", "text", "-E", ".", "_build"])

    with open(
        os.path.join(doc_folder_with_code, "_build", "index.txt"), "rb"
    ) as f:
        content = _sanitise_value(f.read())

        assert content == (
            "example.doSomething1(arg1, arg2 = null)\n"
            "\n"
            "   \"import {doSomething1} from \"example\"\"\n"
            "\n"
            "   A function\n"
            "\n"
            "   Note: A note.\n"
            "\n"
            "example.doSomething2(arg)\n"
            "\n"
            "   Another function\n"
            "\n"
            "   A citation:\n"
            "\n"
            "      A citation\n"
            "\n"
            "example.doSomething3()\n"
            "\n"
            "   \"import {doSomething3} from \"example\"\"\n"
            "\n"
            "example.__ANONYMOUS_FUNCTION__()\n"
            "\n"
            "function* example.yieldSomethingAliased(arg)\n"
            "\n"
            "   generator function\n"
        )



# :coding: utf-8

from sphinx.domains.javascript import JSObject

from .rst_generator import rst_string


class BaseDirective(JSObject):
    """Base class for :term:`Javascript` object directive.
    """

    #: Only the object id argument is required
    required_arguments = 1

    #: No optional argument is available
    optional_arguments = 0

    #: Content is automatically generated and can not be manually entered
    has_content = False

    #: Nested element is automatically generated and can not be manually entered
    allow_nesting = False

    #: No prefix is displayed right before the documentation entry
    display_prefix = None

    def run(self):
        """Run the directive."""
        # The signature is always the first argument.
        signature = self.arguments[0]

        js_env = self.state.document.settings.env.app.config.js_environment
        if signature not in js_env[self.objtype].keys():
            raise self.error(
                "The {objtype} id is unavailable: {signature}".format(
                    objtype=self.objtype,
                    signature=signature
                )
            )

        env = js_env[self.objtype][signature]
        module_env = js_env["module"]

        # Update references
        ref_context = self.state.document.settings.env.ref_context
        ref_context["js:module"] = env["module_id"]

        # Update settings environment
        self.state.document.settings.env.element_environment = env
        self.state.document.settings.env.module_environment = module_env

        return super(BaseDirective, self).run()

    def generate_import_statement(
        self, environment, module_environment, force_partial_import=False
    ):
        """Return import statement generated from *environment* and
        *module_environment*.

        The import statement will be generated only if the element is exported
        and the usage of partial import will depend whether the element is
        exported as default.

        *force_partial_import* indicate whether the usage of partial import
        should be used even if the element is exported as default.

        .. warning::

            The statement is using :term:`Javascript` ES6
            :js:external:`import <Statements/import>` keyword

        Example:

        .. code-block:: js

            import element from "module"
            import {partialElement} from "module"

        """
        name = self.options.get("alias", environment["name"])
        module_id = environment["module_id"]
        module_path = self.options.get(
            "module-path-alias", module_environment[module_id]["path"]
        )
        exported = environment["exported"]
        is_default = environment["default"]

        if exported and is_default and not force_partial_import:
            return rst_string(
                "``import {name} from \"{module}\"``\n".format(
                    name=name, module=module_path
                )
            )

        if exported and (not is_default or force_partial_import):
            return rst_string(
                "``import {{{name}}} from \"{module}\"``\n".format(
                    name=name, module=module_path
                )
            )

        return rst_string()

    def generate_description(self, environment):
        """Return description generated from *environment*.
        """
        description = environment["description"]

        content = rst_string()
        if description:
            content += rst_string(description)

        return content



# :coding: utf-8

import pytest
import os
import re
import unicodedata

from sphinx.cmdline import main as sphinx_main
from sphinx.util.osutil import cd


def _sanitise_value(value):
    """Return *value* suitable for comparison using python 2 and python 3.
    """
    value = value.decode("UTF-8")
    value = unicodedata.normalize("NFKD", value)
    value = value.encode("ascii", "ignore").decode("UTF-8")
    value = re.sub(
        r"[^\w._\-\\/:% \"()\[\]{}\n=,]", "", value
    )
    return value


@pytest.fixture()
def doc_folder_with_code(doc_folder):
    """Return Doc folder with Javascript example source code.
    """
    js_source = os.path.join(doc_folder, "example")

    with open(os.path.join(js_source, "index.js"), "w") as f:
        f.write(
            "/**\n"
            " * Base Class\n"
            " */\n"
            "class MotherClass {\n"
            "    constructor() {\n"
            "        this.attribute = 42\n"
            "    }\n"
            "}\n"
            "\n"
            "const CustomWelcome = class Welcome {\n"
            "    greeting() {\n"
            "        return 'Hello World';\n"
            "    }\n"
            "};\n"
            "\n"
            "/**\n"
            " * Inherited class\n"
            " */\n"
            "export default class AwesomeClass extends MotherClass {\n"
            "\n"
            "    /**\n"
            "     * Constructor.\n"
            "     */\n"
            "    constructor(name) {\n"
            "        super();\n"
            "        this.name = name;\n"
            "    }\n"
            "\n"
            "    /**\n"
            "     * Get name.\n"
            "     *\n"
            "     * .. warning::\n"
            "     *\n"
            "     *     The name is awesome\n"
            "     */\n"
            "    get name() {\n"
            "        return this.name;\n"
            "    }\n"
            "\n"
            "    /**\n"
            "     * Set name.\n"
            "     *\n"
            "     * .. warning::\n"
            "     *\n"
            "     *     Keep the name awesome\n"
            "     */\n"
            "    set name(value) {\n"
            "        this.name = value;\n"
            "    }\n"
            "\n"
            "    /**\n"
            "     * awesomeMethod.\n"
            "     */\n"
            "    awesomeMethod = () => {\n"
            "        console.log('Method has been called');\n"
            "    };\n"
            "\n"
            "    undocumentedMethod(arg1, arg2) {\n"
            "        console.log('An un-documented method has been called');\n"
            "    }\n"
            "\n"
            "    /**\n"
            "     * Private Method.\n"
            "     */\n"
            "    _privateMethod(arg1) {\n"
            "        console.log('An private method has been called');\n"
            "    }\n"
            "\n"
            "    /**\n"
            "     * staticMethod.\n"
            "     */\n"
            "    static staticMethod() {\n"
            "        console.log('Static method has been called');\n"
            "    }\n"
            "\n"
            "    /**\n"
            "     * attribute.\n"
            "     */\n"
            "    static attribute = 42;\n"
            "\n"
            "    /**\n"
            "     * another attribute.\n"
            "     */\n"
            "    classicAttribute = {\n"
            "        test: 'a test',\n"
            "    };\n"
            "}\n"
        )

    return doc_folder


def test_directive_autoclass(doc_folder_with_code):
    """Generate documentation from classes.
    """
    index_file = os.path.join(doc_folder_with_code, "index.rst")
    with open(index_file, "w") as f:
        f.write(
            ".. js:autoclass:: example.MotherClass\n"
            "\n"
            ".. js:autoclass:: example.CustomWelcome\n"
            "\n"
            ".. js:autoclass:: example.AwesomeClass\n"
        )

    with cd(doc_folder_with_code):
        sphinx_main(["-c", ".", "-b", "text", "-E", ".", "_build"])

    with open(
        os.path.join(doc_folder_with_code, "_build", "index.txt"), "rb"
    ) as f:
        content = _sanitise_value(f.read())

        assert content == (
            "class example.MotherClass()\n"
            "\n"
            "   Base Class\n"
            "\n"
            "class example.CustomWelcome()\n"
            "\n"
            "class example.AwesomeClass(name)\n"
            "\n"
            "   \"import AwesomeClass from \"example\"\"\n"
            "\n"
            "   Inherited class\n"
        )


def test_directive_autoclass_with_members(doc_folder_with_code):
    """Generate documentation from classes with all members.
    """
    index_file = os.path.join(doc_folder_with_code, "index.rst")
    with open(index_file, "w") as f:
        f.write(
            ".. js:autoclass:: example.MotherClass\n"
            "    :members:\n"
            "\n"
            ".. js:autoclass:: example.CustomWelcome\n"
            "    :members:\n"
            "\n"
            ".. js:autoclass:: example.AwesomeClass\n"
            "    :members:\n"
        )

    with cd(doc_folder_with_code):
        sphinx_main(["-c", ".", "-b", "text", "-E", ".", "_build"])

    with open(
        os.path.join(doc_folder_with_code, "_build", "index.txt"), "rb"
    ) as f:
        content = _sanitise_value(f.read())

        assert content == (
            "class example.MotherClass()\n"
            "\n"
            "   Base Class\n"
            "\n"
            "class example.CustomWelcome()\n"
            "\n"
            "class example.AwesomeClass(name)\n"
            "\n"
            "   \"import AwesomeClass from \"example\"\"\n"
            "\n"
            "   Inherited class\n"
            "\n"
            "   constructor(name)\n"
            "\n"
            "      Constructor.\n"
            "\n"
            "   get name()\n"
            "\n"
            "      Get name.\n"
            "\n"
            "      Warning: The name is awesome\n"
            "\n"
            "   set name(value)\n"
            "\n"
            "      Set name.\n"
            "\n"
            "      Warning: Keep the name awesome\n"
            "\n"
            "   awesomeMethod()\n"
            "\n"
            "      awesomeMethod.\n"
            "\n"
            "   static staticMethod()\n"
            "\n"
            "      staticMethod.\n"
            "\n"
            "   static attribute = 42\n"
            "\n"
            "      attribute.\n"
            "\n"
            "   classicAttribute = { test: a test, }\n"
            "\n"
            "      another attribute.\n"
        )


def test_directive_autoclass_with_specific_members(doc_folder_with_code):
    """Generate documentation from classes with specific members.
    """
    index_file = os.path.join(doc_folder_with_code, "index.rst")
    with open(index_file, "w") as f:
        f.write(
            ".. js:autoclass:: example.AwesomeClass\n"
            "    :members: name, awesomeMethod\n"
        )

    with cd(doc_folder_with_code):
        sphinx_main(["-c", ".", "-b", "text", "-E", ".", "_build"])

    with open(
        os.path.join(doc_folder_with_code, "_build", "index.txt"), "rb"
    ) as f:
        content = _sanitise_value(f.read())

        assert content == (
            "class example.AwesomeClass(name)\n"
            "\n"
            "   \"import AwesomeClass from \"example\"\"\n"
            "\n"
            "   Inherited class\n"
            "\n"
            "   get name()\n"
            "\n"
            "      Get name.\n"
            "\n"
            "      Warning: The name is awesome\n"
            "\n"
            "   set name(value)\n"
            "\n"
            "      Set name.\n"
            "\n"
            "      Warning: Keep the name awesome\n"
            "\n"
            "   awesomeMethod()\n"
            "\n"
            "      awesomeMethod.\n"
        )


def test_directive_autoclass_with_default_members(doc_folder_with_code):
    """Generate documentation from classes with all members by default.
    """
    conf_file = os.path.join(doc_folder_with_code, "conf.py")
    with open(conf_file, "a") as f:
        f.write("\njs_class_options = ['members']")

    index_file = os.path.join(doc_folder_with_code, "index.rst")
    with open(index_file, "w") as f:
        f.write(
            ".. js:autoclass:: example.MotherClass\n"
            "\n"
            ".. js:autoclass:: example.CustomWelcome\n"
            "\n"
            ".. js:autoclass:: example.AwesomeClass\n"
        )

    with cd(doc_folder_with_code):
        sphinx_main(["-c", ".", "-b", "text", "-E", ".", "_build"])

    with open(
        os.path.join(doc_folder_with_code, "_build", "index.txt"), "rb"
    ) as f:
        content = _sanitise_value(f.read())

        assert content == (
            "class example.MotherClass()\n"
            "\n"
            "   Base Class\n"
            "\n"
            "class example.CustomWelcome()\n"
            "\n"
            "class example.AwesomeClass(name)\n"
            "\n"
            "   \"import AwesomeClass from \"example\"\"\n"
            "\n"
            "   Inherited class\n"
            "\n"
            "   constructor(name)\n"
            "\n"
            "      Constructor.\n"
            "\n"
            "   get name()\n"
            "\n"
            "      Get name.\n"
            "\n"
            "      Warning: The name is awesome\n"
            "\n"
            "   set name(value)\n"
            "\n"
            "      Set name.\n"
            "\n"
            "      Warning: Keep the name awesome\n"
            "\n"
            "   awesomeMethod()\n"
            "\n"
            "      awesomeMethod.\n"
            "\n"
            "   static staticMethod()\n"
            "\n"
            "      staticMethod.\n"
            "\n"
            "   static attribute = 42\n"
            "\n"
            "      attribute.\n"
            "\n"
            "   classicAttribute = { test: a test, }\n"
            "\n"
            "      another attribute.\n"
        )


def test_directive_autoclass_without_constructor(doc_folder_with_code):
    """Generate documentation from classes without constructors.
    """
    index_file = os.path.join(doc_folder_with_code, "index.rst")
    with open(index_file, "w") as f:
        f.write(
            ".. js:autoclass:: example.AwesomeClass\n"
            "    :members:\n"
            "    :skip-constructor:\n"
        )

    with cd(doc_folder_with_code):
        sphinx_main(["-c", ".", "-b", "text", "-E", ".", "_build"])

    with open(
        os.path.join(doc_folder_with_code, "_build", "index.txt"), "rb"
    ) as f:
        content = _sanitise_value(f.read())

        assert content == (
            "class example.AwesomeClass(name)\n"
            "\n"
            "   \"import AwesomeClass from \"example\"\"\n"
            "\n"
            "   Inherited class\n"
            "\n"
            "   get name()\n"
            "\n"
            "      Get name.\n"
            "\n"
            "      Warning: The name is awesome\n"
            "\n"
            "   set name(value)\n"
            "\n"
            "      Set name.\n"
            "\n"
            "      Warning: Keep the name awesome\n"
            "\n"
            "   awesomeMethod()\n"
            "\n"
            "      awesomeMethod.\n"
            "\n"
            "   static staticMethod()\n"
            "\n"
            "      staticMethod.\n"
            "\n"
            "   static attribute = 42\n"
            "\n"
            "      attribute.\n"
            "\n"
            "   classicAttribute = { test: a test, }\n"
            "\n"
            "      another attribute.\n"
        )


def test_directive_autoclass_without_constructor_default(doc_folder_with_code):
    """Generate documentation from classes without constructors by default.
    """
    conf_file = os.path.join(doc_folder_with_code, "conf.py")
    with open(conf_file, "a") as f:
        f.write("\njs_class_options = ['skip-constructor']")

    index_file = os.path.join(doc_folder_with_code, "index.rst")
    with open(index_file, "w") as f:
        f.write(
            ".. js:autoclass:: example.AwesomeClass\n"
            "    :members:\n"
        )

    with cd(doc_folder_with_code):
        sphinx_main(["-c", ".", "-b", "text", "-E", ".", "_build"])

    with open(
        os.path.join(doc_folder_with_code, "_build", "index.txt"), "rb"
    ) as f:
        content = _sanitise_value(f.read())

        assert content == (
            "class example.AwesomeClass(name)\n"
            "\n"
            "   \"import AwesomeClass from \"example\"\"\n"
            "\n"
            "   Inherited class\n"
            "\n"
            "   get name()\n"
            "\n"
            "      Get name.\n"
            "\n"
            "      Warning: The name is awesome\n"
            "\n"
            "   set name(value)\n"
            "\n"
            "      Set name.\n"
            "\n"
            "      Warning: Keep the name awesome\n"
            "\n"
            "   awesomeMethod()\n"
            "\n"
            "      awesomeMethod.\n"
            "\n"
            "   static staticMethod()\n"
            "\n"
            "      staticMethod.\n"
            "\n"
            "   static attribute = 42\n"
            "\n"
            "      attribute.\n"
            "\n"
            "   classicAttribute = { test: a test, }\n"
            "\n"
            "      another attribute.\n"
        )


def test_directive_autoclass_with_undocumented_members(doc_folder_with_code):
    """Generate documentation from classes with undocumented members.
    """
    index_file = os.path.join(doc_folder_with_code, "index.rst")
    with open(index_file, "w") as f:
        f.write(
            ".. js:autoclass:: example.MotherClass\n"
            "    :members:\n"
            "    :undoc-members:\n"
            "\n"
            ".. js:autoclass:: example.CustomWelcome\n"
            "    :members:\n"
            "    :undoc-members:\n"
            "\n"
            ".. js:autoclass:: example.AwesomeClass\n"
            "    :members:\n"
            "    :undoc-members:\n"
        )

    with cd(doc_folder_with_code):
        sphinx_main(["-c", ".", "-b", "text", "-E", ".", "_build"])

    with open(
        os.path.join(doc_folder_with_code, "_build", "index.txt"), "rb"
    ) as f:
        content = _sanitise_value(f.read())

        assert content == (
            "class example.MotherClass()\n"
            "\n"
            "   Base Class\n"
            "\n"
            "   constructor()\n"
            "\n"
            "class example.CustomWelcome()\n"
            "\n"
            "   greeting()\n"
            "\n"
            "class example.AwesomeClass(name)\n"
            "\n"
            "   \"import AwesomeClass from \"example\"\"\n"
            "\n"
            "   Inherited class\n"
            "\n"
            "   constructor(name)\n"
            "\n"
            "      Constructor.\n"
            "\n"
            "   get name()\n"
            "\n"
            "      Get name.\n"
            "\n"
            "      Warning: The name is awesome\n"
            "\n"
            "   set name(value)\n"
            "\n"
            "      Set name.\n"
            "\n"
            "      Warning: Keep the name awesome\n"
            "\n"
            "   awesomeMethod()\n"
            "\n"
            "      awesomeMethod.\n"
            "\n"
            "   undocumentedMethod(arg1, arg2)\n"
            "\n"
            "   static staticMethod()\n"
            "\n"
            "      staticMethod.\n"
            "\n"
            "   static attribute = 42\n"
            "\n"
            "      attribute.\n"
            "\n"
            "   classicAttribute = { test: a test, }\n"
            "\n"
            "      another attribute.\n"
        )


def test_directive_autoclass_with_undoc_members_default(doc_folder_with_code):
    """Generate documentation from classes with undocumented members
    by default.
    """
    conf_file = os.path.join(doc_folder_with_code, "conf.py")
    with open(conf_file, "a") as f:
        f.write("\njs_class_options = ['undoc-members']")

    index_file = os.path.join(doc_folder_with_code, "index.rst")
    with open(index_file, "w") as f:
        f.write(
            ".. js:autoclass:: example.MotherClass\n"
            "    :members:\n"
            "\n"
            ".. js:autoclass:: example.CustomWelcome\n"
            "    :members:\n"
            "\n"
            ".. js:autoclass:: example.AwesomeClass\n"
            "    :members:\n"
        )

    with cd(doc_folder_with_code):
        sphinx_main(["-c", ".", "-b", "text", "-E", ".", "_build"])

    with open(
        os.path.join(doc_folder_with_code, "_build", "index.txt"), "rb"
    ) as f:
        content = _sanitise_value(f.read())

        assert content == (
            "class example.MotherClass()\n"
            "\n"
            "   Base Class\n"
            "\n"
            "   constructor()\n"
            "\n"
            "class example.CustomWelcome()\n"
            "\n"
            "   greeting()\n"
            "\n"
            "class example.AwesomeClass(name)\n"
            "\n"
            "   \"import AwesomeClass from \"example\"\"\n"
            "\n"
            "   Inherited class\n"
            "\n"
            "   constructor(name)\n"
            "\n"
            "      Constructor.\n"
            "\n"
            "   get name()\n"
            "\n"
            "      Get name.\n"
            "\n"
            "      Warning: The name is awesome\n"
            "\n"
            "   set name(value)\n"
            "\n"
            "      Set name.\n"
            "\n"
            "      Warning: Keep the name awesome\n"
            "\n"
            "   awesomeMethod()\n"
            "\n"
            "      awesomeMethod.\n"
            "\n"
            "   undocumentedMethod(arg1, arg2)\n"
            "\n"
            "   static staticMethod()\n"
            "\n"
            "      staticMethod.\n"
            "\n"
            "   static attribute = 42\n"
            "\n"
            "      attribute.\n"
            "\n"
            "   classicAttribute = { test: a test, }\n"
            "\n"
            "      another attribute.\n"
        )


def test_directive_autoclass_with_private_members(doc_folder_with_code):
    """Generate documentation from classes with private members.
    """
    index_file = os.path.join(doc_folder_with_code, "index.rst")
    with open(index_file, "w") as f:
        f.write(
            ".. js:autoclass:: example.AwesomeClass\n"
            "    :members:\n"
            "    :private-members:\n"
        )

    with cd(doc_folder_with_code):
        sphinx_main(["-c", ".", "-b", "text", "-E", ".", "_build"])

    with open(
        os.path.join(doc_folder_with_code, "_build", "index.txt"), "rb"
    ) as f:
        content = _sanitise_value(f.read())

        assert content == (
            "class example.AwesomeClass(name)\n"
            "\n"
            "   \"import AwesomeClass from \"example\"\"\n"
            "\n"
            "   Inherited class\n"
            "\n"
            "   constructor(name)\n"
            "\n"
            "      Constructor.\n"
            "\n"
            "   get name()\n"
            "\n"
            "      Get name.\n"
            "\n"
            "      Warning: The name is awesome\n"
            "\n"
            "   set name(value)\n"
            "\n"
            "      Set name.\n"
            "\n"
            "      Warning: Keep the name awesome\n"
            "\n"
            "   awesomeMethod()\n"
            "\n"
            "      awesomeMethod.\n"
            "\n"
            "   _privateMethod(arg1)\n"
            "\n"
            "      Private Method.\n"
            "\n"
            "   static staticMethod()\n"
            "\n"
            "      staticMethod.\n"
            "\n"
            "   static attribute = 42\n"
            "\n"
            "      attribute.\n"
            "\n"
            "   classicAttribute = { test: a test, }\n"
            "\n"
            "      another attribute.\n"
        )


def test_directive_autoclass_with_private_members_default(doc_folder_with_code):
    """Generate documentation from classes with private members by default.
    """
    conf_file = os.path.join(doc_folder_with_code, "conf.py")
    with open(conf_file, "a") as f:
        f.write("\njs_class_options = ['private-members']")

    index_file = os.path.join(doc_folder_with_code, "index.rst")
    with open(index_file, "w") as f:
        f.write(
            ".. js:autoclass:: example.AwesomeClass\n"
            "    :members:\n"
        )

    with cd(doc_folder_with_code):
        sphinx_main(["-c", ".", "-b", "text", "-E", ".", "_build"])

    with open(
        os.path.join(doc_folder_with_code, "_build", "index.txt"), "rb"
    ) as f:
        content = _sanitise_value(f.read())

        assert content == (
            "class example.AwesomeClass(name)\n"
            "\n"
            "   \"import AwesomeClass from \"example\"\"\n"
            "\n"
            "   Inherited class\n"
            "\n"
            "   constructor(name)\n"
            "\n"
            "      Constructor.\n"
            "\n"
            "   get name()\n"
            "\n"
            "      Get name.\n"
            "\n"
            "      Warning: The name is awesome\n"
            "\n"
            "   set name(value)\n"
            "\n"
            "      Set name.\n"
            "\n"
            "      Warning: Keep the name awesome\n"
            "\n"
            "   awesomeMethod()\n"
            "\n"
            "      awesomeMethod.\n"
            "\n"
            "   _privateMethod(arg1)\n"
            "\n"
            "      Private Method.\n"
            "\n"
            "   static staticMethod()\n"
            "\n"
            "      staticMethod.\n"
            "\n"
            "   static attribute = 42\n"
            "\n"
            "      attribute.\n"
            "\n"
            "   classicAttribute = { test: a test, }\n"
            "\n"
            "      another attribute.\n"
        )


def test_directive_autoclass_with_alias(doc_folder_with_code):
    """Generate documentation from classes with alias.
    """
    index_file = os.path.join(doc_folder_with_code, "index.rst")
    with open(index_file, "w") as f:
        f.write(
            ".. js:autoclass:: example.MotherClass\n"
            "    :alias: AliasedMotherClass\n"
            "\n"
            ".. js:autoclass:: example.CustomWelcome\n"
            "    :alias: AliasedCustomWelcome\n"
            "\n"
            ".. js:autoclass:: example.AwesomeClass\n"
            "    :alias: AliasedAwesomeClass\n"
        )

    with cd(doc_folder_with_code):
        sphinx_main(["-c", ".", "-b", "text", "-E", ".", "_build"])

    with open(
        os.path.join(doc_folder_with_code, "_build", "index.txt"), "rb"
    ) as f:
        content = _sanitise_value(f.read())

        assert content == (
            "class example.AliasedMotherClass()\n"
            "\n"
            "   Base Class\n"
            "\n"
            "class example.AliasedCustomWelcome()\n"
            "\n"
            "class example.AliasedAwesomeClass(name)\n"
            "\n"
            "   \"import AliasedAwesomeClass from \"example\"\"\n"
            "\n"
            "   Inherited class\n"
        )


def test_directive_autoclass_with_module_alias(doc_folder_with_code):
    """Generate documentation from classes with module alias.
    """
    index_file = os.path.join(doc_folder_with_code, "index.rst")
    with open(index_file, "w") as f:
        f.write(
            ".. js:autoclass:: example.MotherClass\n"
            "    :module-alias: alias_module\n"
            "\n"
            ".. js:autoclass:: example.CustomWelcome\n"
            "    :module-alias: alias_module\n"
            "\n"
            ".. js:autoclass:: example.AwesomeClass\n"
            "    :module-alias: alias_module\n"
        )

    with cd(doc_folder_with_code):
        sphinx_main(["-c", ".", "-b", "text", "-E", ".", "_build"])

    with open(
        os.path.join(doc_folder_with_code, "_build", "index.txt"), "rb"
    ) as f:
        content = _sanitise_value(f.read())

        assert content == (
            "class alias_module.MotherClass()\n"
            "\n"
            "   Base Class\n"
            "\n"
            "class alias_module.CustomWelcome()\n"
            "\n"
            "class alias_module.AwesomeClass(name)\n"
            "\n"
            "   \"import AwesomeClass from \"example\"\"\n"
            "\n"
            "   Inherited class\n"
        )


def test_directive_autoclass_with_module_path_alias(doc_folder_with_code):
    """Generate documentation from classes with module path alias.
    """
    index_file = os.path.join(doc_folder_with_code, "index.rst")
    with open(index_file, "w") as f:
        f.write(
            ".. js:autoclass:: example.MotherClass\n"
            "    :module-path-alias: test/alias/module\n"
            "\n"
            ".. js:autoclass:: example.CustomWelcome\n"
            "    :module-path-alias: test/alias/module\n"
            "\n"
            ".. js:autoclass:: example.AwesomeClass\n"
            "    :module-path-alias: test/alias/module\n"
        )

    with cd(doc_folder_with_code):
        sphinx_main(["-c", ".", "-b", "text", "-E", ".", "_build"])

    with open(
        os.path.join(doc_folder_with_code, "_build", "index.txt"), "rb"
    ) as f:
        content = _sanitise_value(f.read())

        assert content == (
            "class example.MotherClass()\n"
            "\n"
            "   Base Class\n"
            "\n"
            "class example.CustomWelcome()\n"
            "\n"
            "class example.AwesomeClass(name)\n"
            "\n"
            "   \"import AwesomeClass from \"test/alias/module\"\"\n"
            "\n"
            "   Inherited class\n"
        )


def test_directive_autoclass_with_partial_import_forced(doc_folder_with_code):
    """Generate documentation from classes with partial import forced.
    """
    index_file = os.path.join(doc_folder_with_code, "index.rst")
    with open(index_file, "w") as f:
        f.write(
            ".. js:autoclass:: example.AwesomeClass\n"
            "    :force-partial-import:\n"
        )

    with cd(doc_folder_with_code):
        sphinx_main(["-c", ".", "-b", "text", "-E", ".", "_build"])

    with open(
        os.path.join(doc_folder_with_code, "_build", "index.txt"), "rb"
    ) as f:
        content = _sanitise_value(f.read())

        assert content == (
            "class example.AwesomeClass(name)\n"
            "\n"
            "   \"import {AwesomeClass} from \"example\"\"\n"
            "\n"
            "   Inherited class\n"
        )


def test_directive_autoclass_with_attribute_value_skipped(doc_folder_with_code):
    """Generate documentation from classes with data value skipped.
    """
    index_file = os.path.join(doc_folder_with_code, "index.rst")
    with open(index_file, "w") as f:
        f.write(
            ".. js:autoclass:: example.AwesomeClass\n"
            "    :members:\n"
            "    :skip-attribute-value:\n"
        )

    with cd(doc_folder_with_code):
        sphinx_main(["-c", ".", "-b", "text", "-E", ".", "_build"])

    with open(
        os.path.join(doc_folder_with_code, "_build", "index.txt"), "rb"
    ) as f:
        content = _sanitise_value(f.read())

        assert content == (
            "class example.AwesomeClass(name)\n"
            "\n"
            "   \"import AwesomeClass from \"example\"\"\n"
            "\n"
            "   Inherited class\n"
            "\n"
            "   constructor(name)\n"
            "\n"
            "      Constructor.\n"
            "\n"
            "   get name()\n"
            "\n"
            "      Get name.\n"
            "\n"
            "      Warning: The name is awesome\n"
            "\n"
            "   set name(value)\n"
            "\n"
            "      Set name.\n"
            "\n"
            "      Warning: Keep the name awesome\n"
            "\n"
            "   awesomeMethod()\n"
            "\n"
            "      awesomeMethod.\n"
            "\n"
            "   static staticMethod()\n"
            "\n"
            "      staticMethod.\n"
            "\n"
            "   static attribute\n"
            "\n"
            "      attribute.\n"
            "\n"
            "   classicAttribute\n"
            "\n"
            "      another attribute.\n"
        )



# :coding: utf-8

import re
import functools

from .helper import collapse_all
from .helper import get_docstring


#: Regular Expression pattern for data
_DATA_PATTERN = re.compile(
    r"(?P<start_regex>(\n|^)) *(?P<export>export +)?(?P<default>default +)?"
    r"(?P<type>(const|let|var)) (?P<name>[\w._-]+) *= *(?P<value>.+?;)",
    re.DOTALL
)


def fetch_environment(content, module_id):
    """Return data environment dictionary from *content*.

    *module_id* represent the identifier of the module.

    The environment is in the form of::

        {
            "moduleName.DATA": {
                "id": "moduleName.DATA",
                "module_id": "moduleName",
                "exported": False,
                "default": False,
                "name": "DATA",
                "value": "42",
                "type": "const",
                "line_number": 2,
                "description": "Variable doc.\\n\\nDetailed description."
            }
        }

    """
    environment = {}

    lines = content.split("\n")

    # The comment filter is made during the collapse content process to
    # preserve the entire value (with semi-colons and docstrings!)
    content, collapsed_content = collapse_all(content, filter_comment=True)

    for match in _DATA_PATTERN.finditer(content):
        data_id = ".".join([module_id, match.group("name")])

        line_number = (
            content[:match.start()].count("\n") +
            match.group("start_regex").count("\n") + 1
        )

        value = match.group("value")
        if "{}" in value and line_number in collapsed_content.keys():
            value = value.replace("{}", collapsed_content[line_number])

        # Do not keep semi-colon in value
        if value.endswith(";"):
            value = value[:-1]

        data_environment = {
            "id": data_id,
            "module_id": module_id,
            "exported": match.group("export") is not None,
            "default": match.group("default") is not None,
            "name": match.group("name"),
            "value": functools.reduce(_clean_value, value.split('\n')).strip(),
            "type": match.group("type"),
            "line_number": line_number,
            "description": get_docstring(line_number, lines)
        }
        environment[data_id] = data_environment

    return environment


def _clean_value(line1, line2):
    """Clean up variable value for display."""
    _line1 = line1.strip()
    _line2 = line2.strip()

    # Let trailing space to make the code easier to read
    if _line1[-1:] in ["{", "}", "(", ")", "[", "]", ";", ","]:
        _line1 += " "

    return _line1 + _line2



# :coding: utf-8

import collections

from sphinx import addnodes
import docutils.parsers.rst.directives

from .base import BaseDirective

from .rst_generator import (
    get_rst_attribute_elements,
    get_rst_method_elements
)


def _parse_members(argument):
    """Convert the :members: options to class directive."""
    if argument is None:
        return True
    return [arg.strip() for arg in argument.split(",")]


class AutoClassDirective(BaseDirective):
    """Directive to render :term:`Javascript` class documentation.

    The unique argument should be the identifier of the class element.

    .. sourcecode:: rest

        .. js:autoclass:: module.AwesomeClass

    The available options are:

    * members:
        This option can be boolean if no arguments are given to indicate that
        all members should be documented, or a white list of member names to
        display.

    * skip-constructor:
        Indicate whether the constructor method should be displayed if
        available.

    * skip-attribute-value:
        Indicate whether attribute values within the class should be skipped.

    * undoc-members:
        Indicate whether members with no docstrings should be displayed.

    * private-members:
        Indicate whether private members (with a name starting with an
        underscore) should be displayed.

    * alias:
        String element to replace the class name.

    * module-alias:
        String element to replace the module name.

    * module-path-alias:
        String element to replace the module path.

    * force-partial-import:
        Indicate whether the class import statement display should be indicated
        with partial import if the class element is exported.

    .. seealso::

        :ref:`directive/autoclass`

    .. seealso::

        :ref:`configuration/js_class_options`

    """
    #: Javascript class is callable
    has_arguments = True

    #: Define the Object type
    objtype = "class"

    #: classes options
    option_spec = {
        "members": _parse_members,
        "skip-constructor": lambda x: True,
        "skip-attribute-value": lambda x: True,
        "undoc-members": lambda x: True,
        "private-members": lambda x: True,
        "alias": docutils.parsers.rst.directives.unchanged_required,
        "module-alias": docutils.parsers.rst.directives.unchanged_required,
        "module-path-alias": docutils.parsers.rst.directives.unchanged_required,
        "force-partial-import": lambda x: True,
    }

    def handle_signature(self, signature, node):
        """Update the signature *node*."""
        env = self.state.document.settings.env.element_environment
        module_env = self.state.document.settings.env.module_environment

        name = self.options.get("alias", env["name"])
        module_id = env["module_id"]
        module_name = self.options.get(
            "module-alias", module_env[module_id]["name"]
        )

        node["type"] = "class"
        node["id"] = env["id"]
        node["module"] = module_name
        node["fullname"] = name

        node += addnodes.desc_type("class ", "class ")
        node += addnodes.desc_addname(module_name + ".", module_name + ".")
        node += addnodes.desc_name(name, name)

        param_list = addnodes.desc_parameterlist()
        for method_environment in env["method"].values():
            if method_environment["name"] == "constructor":
                for argument in method_environment["arguments"]:
                    param_list += addnodes.desc_parameter(argument, argument)
        node += param_list

        return name, module_name

    def before_content(self):
        """Update the content.

        Compute the description and import statement if available, and generate
        the class nested directive elements to integrate to the content.

        """
        env = self.state.document.settings.env.element_environment
        module_env = self.state.document.settings.env.module_environment

        self.content = self.generate_import_statement(
            env, module_env, self.options.get("force-partial-import")
        )
        self.content += self.generate_description(env)

        # Automatic boolean options
        options = self.env.config.js_class_options

        # Options manually set
        skip_constructor = self.options.get(
            "skip-constructor", "skip-constructor" in options
        )
        undoc_members = self.options.get(
            "undoc-members", "undoc-members" in options
        )
        private_members = self.options.get(
            "private-members", "private-members" in options
        )

        members = self.options.get("members", "members" in options)
        if members:
            rst_elements = {}

            whitelist = (
                members if isinstance(members, collections.Iterable) else None
            )

            # Gather class attributes
            rst_elements = get_rst_attribute_elements(
                env,
                whitelist_names=whitelist,
                blacklist_ids=env["method"].keys(),
                undocumented_members=undoc_members,
                private_members=private_members,
                skip_value=self.options.get(
                    "skip-attribute-value", "skip-attribute-value" in options
                ),
                rst_elements=rst_elements
            )

            # Gather class methods
            rst_elements = get_rst_method_elements(
                env,
                whitelist_names=whitelist,
                skip_constructor=skip_constructor,
                undocumented_members=undoc_members,
                private_members=private_members,
                rst_elements=rst_elements,
            )

            # Add content while respecting the line order
            for line_number in sorted(rst_elements.keys()):
                for rst_element in rst_elements[line_number]:
                    self.content += rst_element


class AutoMethodDirective(BaseDirective):
    """Directive to render :term:`Javascript` class method documentation.

    The unique argument should be the identifier of the class method element.

    .. sourcecode:: rest

        .. js:automethod:: module.AwesomeClass.awesomeMethod

    .. seealso::

        :ref:`directive/automethod`

    """
    #: Javascript method is callable
    has_arguments = True

    #: Define the Object type
    objtype = "method"

    def handle_signature(self, signature, node):
        """Update the signature *node*."""
        env = self.state.document.settings.env.element_environment

        name = env["name"]
        prefix = env["prefix"]

        node["type"] = "method"
        node["id"] = env["id"]
        node["fullname"] = name

        if prefix is not None:
            node += addnodes.desc_type(prefix + " ", prefix + " ")
        node += addnodes.desc_name(name, name)

        param_list = addnodes.desc_parameterlist()
        for argument in env["arguments"]:
            param_list += addnodes.desc_parameter(argument, argument)
        node += param_list

        class_name = env["class_id"].rsplit(".", 1)[-1]
        return class_name + "." + name, None

    def before_content(self):
        """Update the content.

        Compute the description if available.

        """
        env = self.state.document.settings.env.element_environment

        self.content = self.generate_description(env)


class AutoAttributeDirective(BaseDirective):
    """Directive to render :term:`Javascript` class attribute documentation.

    The unique argument should be the identifier of the class attribute element.

    .. sourcecode:: rest

        .. js:autoattribute:: module.AwesomeClass.DATA

    The available options are:

    * skip-value:
        Indicate whether attribute value should be skipped.

    .. seealso::

        :ref:`directive/autoattribute`

    """
    #: Javascript data are not callable
    has_arguments = False

    #: Define the Object type
    objtype = "attribute"

    #: data options
    option_spec = {
        "skip-value": lambda x: True,
    }

    def handle_signature(self, signature, node):
        """Update the signature *node*."""
        env = self.state.document.settings.env.element_environment

        name = env["name"]
        value = env["value"]
        prefix = env["prefix"]

        node["type"] = "attribute"
        node["id"] = env["id"]
        node["fullname"] = name

        skip_value = self.options.get("skip-value", False)

        if prefix is not None:
            node += addnodes.desc_type(prefix + " ", prefix + " ")
        node += addnodes.desc_name(name, name)
        if not skip_value:
            node += addnodes.desc_annotation(" = " + value, " = " + value)

        class_name = env["class_id"].rsplit(".", 1)[-1]
        return class_name + "." + name, None

    def before_content(self):
        """Update the content.

        Compute the description if available.

        """
        env = self.state.document.settings.env.element_environment
        self.content = self.generate_description(env)



# :coding: utf-8

import re

from .helper import filter_comments
from .helper import collapse_all
from .helper import get_docstring


#: Regular Expression pattern for function expressions
_FUNCTION_PATTERN = re.compile(
    r"(?P<start_regex>(\n|^)) *(?P<export>export +)?(?P<default>default +)?"
    r"((const|var|let) (?P<data_name>[\w_-]+) *= *)?"
    r"function *(?P<generator>\* *)?(?P<function_name>[\w_-]+)? "
    r"*\([\n ]*(?P<arguments>.*?)[\n ]*\) *{",
    re.DOTALL
)

#: Regular Expression pattern for arrow functions
_FUNCTION_ARROW_PATTERN = re.compile(
    r"(?P<start_regex>(\n|^)) *(?P<export>export +)?(?P<default>default +)?"
    r"(const|let|var) (?P<function_name>\w+) *= *"
    r"(\([\n ]*(?P<arguments>.*?)[\n ]*\)|(?P<single_argument>[\w._-]+)) *"
    r"=> *",
    re.DOTALL
)

#: Regular Expression pattern for imported functions
_IMPORTED_FUNCTION_PATTERN = re.compile(
    r"(?P<start_regex>(\n|^)) *(?P<export>export +)?(?P<default>default +)?"
    r"(?P<function_name>[\w_-]+)? *\([\n ]*(?P<arguments>.*?)[\n ]*\);?",
    re.DOTALL
)


def fetch_environment(content, module_id):
    """Return function environment dictionary from *content*.

    *module_id* represent the identifier of the module.

    The environment is in the form of::

        {
            "moduleName.doSomething": {
                "id": "moduleName.doSomething",
                "module_id": "moduleName",
                "exported": False,
                "default": False,
                "name": "doSomething",
                "anonymous": False,
                "generator": False,
                "arguments": ["argument1", "argument2"],
                "line_number": 2,
                "description": "Function doc.\\n\\nDetailed description."
            }
        }

    """
    environment = {}

    lines = content.split("\n")
    content = filter_comments(content)
    content = collapse_all(content)[0]

    for match_iter in (
        _FUNCTION_ARROW_PATTERN.finditer(content),
        _FUNCTION_PATTERN.finditer(content),
        _IMPORTED_FUNCTION_PATTERN.finditer(content),
    ):
        for match in match_iter:
            name = match.group("function_name")

            # In case of anonymous functions, do not allow the imported
            # function pattern to confuse the 'function' statement with
            # the name.
            if name == "function":
                continue

            if "data_name" in match.groupdict().keys():
                _name = match.group("data_name")
                if _name is not None:
                    name = _name

            generator = False
            if "generator" in match.groupdict().keys():
                generator = match.group("generator") is not None

            if name is None:
                name = "__ANONYMOUS_FUNCTION__"

            function_id = ".".join([module_id, name])

            line_number = (
                content[:match.start()].count("\n") +
                match.group("start_regex").count("\n") + 1
            )

            arguments_matched = match.group("arguments")
            if arguments_matched is None:
                arguments_matched = match.group("single_argument")

            arguments = list(filter(lambda x: len(x), [
                arg.strip() for arg in arguments_matched.split(",")
            ]))

            function_environment = {
                "id": function_id,
                "module_id": module_id,
                "exported": match.group("export") is not None,
                "default": match.group("default") is not None,
                "name": name,
                "anonymous": name == "__ANONYMOUS_FUNCTION__",
                "generator": generator,
                "arguments": arguments,
                "line_number": line_number,
                "description": get_docstring(line_number, lines)
            }
            environment[function_id] = function_environment

    return environment



# :coding: utf-8

"""Parser to fetch_environment all information from a :term:`Javascript` API in order
to document each element from a simple identifier.
"""

import os

from .js_module import fetch_environment as fetch_module_environment
from .js_file import fetch_environment as fetch_file_environment


def fetch_environment(path):
    """Return :term:`Javascript` environment dictionary from *path* structure.

    Raises :exc:`OSError` if the directory is incorrect.

    The environment is in the form of::

        {
            "module": {
                "module.id": {
                    "id": "module.id",
                    "name": "module_name",
                    "file_id": "file/id/index.js"
                    "description": "A module."
                    ...
                },
                ...
            },
            "file": {
                "file/id/index.js": {
                    "id": "file/id/index.js",
                    "module_id": "module_id",
                    "content": "...",
                    ...
                },
                ...
            },
            "class": {
                "class_id": {
                    "id": "class_id",
                    "module_id": "module_id"
                    "description": "A class."
                    ...
                },
                ...
            },
            "method": {
                "method_id": {
                    "id": "method_id",
                    "class_id": "class_id",
                    "module_id": "module_id",
                    "description": "A method."
                    ...
                },
                ...
            },
            "attribute": {
                "attribute_id": {
                    "id": "attribute_id",
                    "class_id": "class_id",
                    "module_id": "module_id",
                    "description": "An attribute."
                    ...
                },
                ...
            },
            "function": {
                "function_id": {
                    "id": "function_id",
                    "module_id": "module_id",
                    "description": "A function."
                    ...
                },
                ...
            },
            "data": {
                "data_id": {
                    "id": "data_id",
                    "module_id": "module_id",
                    "description": "A variable."
                    ...
                },
                ...
            }
        }

    """
    if not os.path.isdir(path) or not os.access(path, os.R_OK):
        raise OSError(
            "The javascript package directory is incorrect: {0}".format(path)
        )

    environment = {
        "module": {},
        "class": {},
        "method": {},
        "attribute": {},
        "function": {},
        "data": {},
        "file": {}
    }

    repository_name = os.path.basename(path)

    extensions = [".js", ".jsx"]

    for root, dirs, files in os.walk(path):
        root_folders = (
            [repository_name] + root.split(path)[-1].split(os.sep)[1:]
        )

        files[:] = [
            f for f in files
            if os.path.splitext(f)[1] in extensions
            and not f.startswith(".")
        ]

        dirs[:] = [
            d for d in dirs
            if not d.startswith(".")
        ]

        for _file in files:
            file_id = "/".join(root_folders + [_file])
            file_path = os.path.join(root, _file)

            # Fetch module environment
            _module_environment = fetch_module_environment(
                file_id, files, module_names=[
                    _module["name"] for _module in
                    environment["module"].values()
                ]
            )
            module_id = _module_environment["id"]
            environment["module"][module_id] = _module_environment

            # Fetch file environment
            _file_environment = fetch_file_environment(
                file_path, file_id, _module_environment["id"]
            )
            file_id = _file_environment["id"]

            method_environment = {}
            attribute_environment = {}

            # Extract methods and attributes from class environment to set it
            # in the top level environment.
            for _class in _file_environment["class"].values():
                method_environment.update(_class["method"].copy())
                attribute_environment.update(_class["attribute"].copy())

            environment["file"][file_id] = _file_environment
            environment["function"].update(_file_environment["function"])
            environment["data"].update(_file_environment["data"])
            environment["class"].update(_file_environment["class"])
            environment["method"].update(method_environment)
            environment["attribute"].update(attribute_environment)

    return environment



# :coding: utf-8

import os
import re

from .js_class import fetch_environment as fetch_class_environment
from .js_function import fetch_environment as fetch_function_environment
from .js_data import fetch_environment as fetch_data_environment

from .helper import get_docstring, filter_comments


#: Regular Expression pattern for imported element
_IMPORTED_ELEMENT_PATTERN = re.compile(
    r"(?P<start_regex>(\n|^)) *import +"
    r"(?P<expression>({([^{}]|\n)+}|.+))"
    r" +from +['\"](?P<module>[\w/.\\_-]+)['\"];?"
)

#: Regular Expression pattern for exported element
_EXPORTED_ELEMENT_PATTERN = re.compile(
    r"(?P<start_regex>(\n|^)) *export +(?P<default>default +)?"
    r"((?P<expression_from_module>({([^{}]|\n)+}|.+))"
    r" +from +['\"](?P<module>[\w/.\\_-]+)['\"]|"
    r"(?P<expression_from_variable>({([^{}]|\n)+}|.+)));?"
)

#: Regular Expression pattern for binding element
_BINDING_ELEMENT_PATTERN = re.compile(
    r"^(?P<name>(\w+|\*))( +as +(?P<alias>\w+))?;?$"
)

#: Regular Expression pattern for file docstring
_FILE_DOCSTRING_PATTERN = re.compile(r"^/\*\*.*?\*/(?=\n\n)", re.DOTALL)


def fetch_environment(file_path, file_id, module_id):
    """Return file environment dictionary from *file_path*.

    *file_id* represent the identifier of the file.

    *module_id* represent the identifier of the module.

    Update the *environment* if available and return it as-is if the file
    is not readable.

    The environment is in the form of::

        {
            "id": "module/test/index.js",
            "module_id": "module.test",
            "name": "index.js",
            "path": "/path/to/module/test/index.js",
            "content": "'use strict'\\n\\n...",
            "description": "File description",
            "export": {
                "module.test.exported_element": {
                    "id": "module.test.exported_element",
                    "module": "module.test.from_module",
                    "description": "An exported element",
                    ...
                },
                ...
            },
            "import": {
                "module.test.imported_element": {
                    "id": "module.test.imported_element",
                    "module": "module.test.from_module",
                    ...
                },
                ...
            },
            "class": {
                "class_id": {
                    "id": "class_id",
                    "module_id": "module_id"
                    "description": "A class."
                    ...
                },
                ...
            },
            "data": {
                "data_id": {
                    "id": "data_id",
                    "module_id": "module_id",
                    "description": "A variable."
                    ...
                },
                ...
            },
            "function": {
                "function_id": {
                    "id": "function_id",
                    "module_id": "module_id",
                    "description": "A function."
                    ...
                },
                ...
            }
        }

    """
    try:
        with open(file_path, "r") as f:
            content = f.read()
    except (IOError, OSError):
        return

    environment = {
        "id": file_id,
        "module_id": module_id,
        "name": os.path.basename(file_path),
        "path": file_path,
        "content": content,
        "description": fetch_file_description(content),
        "export": fetch_export_environment(content, module_id),
        "import": fetch_import_environment(content, module_id),
        "class": {},
        "data": {},
        "function": {}
    }

    for _env_id, _env in fetch_class_environment(content, module_id).items():
        update_from_exported_elements(_env, environment["export"])
        environment["class"][_env_id] = _env

    for _env_id, _env in fetch_function_environment(content, module_id).items():
        update_from_exported_elements(_env, environment["export"])
        environment["function"][_env_id] = _env

    for _env_id, _env in fetch_data_environment(content, module_id).items():
        update_from_exported_elements(_env, environment["export"])
        environment["data"][_env_id] = _env

    return environment


def fetch_file_description(content):
    """Return file description from *content*.

    The description must be in a docstring which should be defined at the very
    beginning of the file. It can only be preceded by one line comments.

    It must be in the form of::

        /**
         * File description.
         *
         * A detailed description of the file.
         *
         */

    Return None if no description is available.

    """
    content = filter_comments(content, filter_multiline_comment=False).strip()

    match = _FILE_DOCSTRING_PATTERN.search(content)
    if match is None:
        return

    docstring_content = match.group()[3:-2].strip()

    # If the entire docstring fit in one line
    if match.group().count("\n") == 0:
        return docstring_content

    docstring = []

    for line in docstring_content.split("\n"):
        line = line.strip()

        # Valid docstring line starts with a '*'
        if re.search("^\*( *| +.+)$", line) is not None:
            indentation = 2 if len(line) > 1 else 1
            docstring.append(line[indentation:].rstrip())

        # Error in the docstring
        else:
            return

    return "\n".join(docstring)


def update_from_exported_elements(environment, export_environment):
    """Update *environment* with exported elements from *export_environment*.

    For instance, the element environment might not be exported, but an
    exported element is found that can be linked to this element.

    .. code-block:: js

        // Element in the environment
        function doSomething(arg1, arg2) {
            console.log("Hello World")
        }

        // Element in the export environment
        export {doSomething};

    In the example above, the function `doSomething` is previously fetched
    without the exported attribute, so it will be added to it.

    The entry will then be removed from the *export_environment*.

    .. warning::

        Both input environments are mutated.

    """
    env_id = environment["id"]

    # Update the environment from exported environment if necessary
    if env_id in export_environment.keys() and not environment["exported"]:
        expression_environment = export_environment[env_id]

        environment["exported"] = True
        environment["default"] = expression_environment["default"]

        if environment["description"] is None:
            environment["description"] = (
                expression_environment["description"]
            )

        # Once the environment is updated, we can remove the key from the
        # exported environment to prevent displaying it twice.
        del export_environment[env_id]


def fetch_import_environment(content, module_id):
    """Return import environment dictionary from *content*.

    *module_id* represent the identifier of the module.

    The environment is in the form of::

        {
            "module.test.imported_element": {
                "id": "module.test.imported_element",
                "module": "module.test.from_module",
                "name": "imported_element",
                "alias": None,
                "partial": False
            },
            ...
        }

    """
    environment = {}

    wildcards_number = 0

    module_path = module_id.replace(".", os.sep)

    for match in _IMPORTED_ELEMENT_PATTERN.finditer(content):
        from_module_path = os.path.normpath(
            os.path.join(module_path, match.group("module"))
        )
        from_module_id = from_module_path.replace(os.sep, ".")

        element_raw = match.group("expression").replace("\n", "")

        _env, wildcards_number = _fetch_expression_environment(
            element_raw, module_id, from_module_id,
            wildcards_number=wildcards_number,
            environment=environment
        )
        environment.update(_env)

    return environment


def fetch_export_environment(content, module_id):
    """Return export environment dictionary from *content*.

    *module_id* represent the identifier of the module.

    The environment is in the form of::

        {
            "module.test.exported_element": {
                "id": "module.test.exported_element",
                "module": "module.test.from_module",
                "name": "exported_element",
                "alias": None,
                "partial": True,
                "description": None,
                "default": False,
                "line_number": 5,
            },
            ...
        }

    """
    environment = {}

    wildcards_number = 0

    lines = content.split("\n")

    module_path = module_id.replace(".", os.sep)

    for match in _EXPORTED_ELEMENT_PATTERN.finditer(content):
        line_number = (
            content[:match.start()].count("\n") +
            match.group("start_regex").count("\n") + 1
        )

        from_module_id = None

        if match.group("module") is not None:
            from_module_path = os.path.normpath(
                os.path.join(module_path, match.group("module"))
            )
            from_module_id = from_module_path.replace(os.sep, ".")

        expression = match.group("expression_from_variable")
        if expression is None:
            expression = match.group("expression_from_module")

        element_raw = expression.replace("\n", "")

        _env, wildcards_number = _fetch_expression_environment(
            element_raw, module_id, from_module_id,
            wildcards_number=wildcards_number,
        )

        for _env_id, _sub_env in _env.items():
            environment[_env_id] = {
                "description": get_docstring(line_number, lines),
                "line_number": line_number,
                "default": match.group("default") is not None,
            }
            environment[_env_id].update(_sub_env)

    return environment


def _fetch_expression_environment(
    expression, module_id, from_module_id=None, environment=None,
    wildcards_number=0
):
    """Return tuple with *expression* environment and updated *wildcards_number*.

    *module_id* represent the identifier of the module.

    *from_module_id* is the optional module identifier from which the
    expression can be resolved.

    Update the *environment* if available and return it as-is if the file
    is not readable.

    *wildcards_number* represent the number of `*` found as un-aliased binding.

    """
    if environment is None:
        environment = {}

    # Parse partial expressions first
    for partial_match in re.finditer("{[^{}]+}", expression):
        binding_environments, wildcards_number = _fetch_binding_environment(
            partial_match.group()[1:-1], module_id, wildcards_number
        )
        for _env in binding_environments:
            element_id = _env["id"]
            environment[element_id] = _env
            environment[element_id].update(
                {"partial": True, "module": from_module_id}
            )

        # remove partial imports from expression
        expression = (
            expression[:partial_match.start()] +
            " " * len(partial_match.group()) +
            expression[partial_match.end():]
        )

    # Then parse the default expressions
    binding_environments, wildcards_number = _fetch_binding_environment(
        expression, module_id, wildcards_number
    )
    for _env in binding_environments:
        element_id = _env["id"]
        environment[element_id] = _env
        environment[element_id].update(
            {"partial": False, "module": from_module_id}
        )

    return environment, wildcards_number


def _fetch_binding_environment(expression, module_id, wildcards_number=0):
    """Return tuple with list of binding environments from *expression* and
    updated *wildcards_number*.

    *module_id* represent the identifier of the module.

    *wildcards_number* represent the number of `*` found as un-aliased binding.

    Example::

        expression = "Module1 as ModuleAlias, Module2"
        env = get_import_module_environment(expression, "test.module")

    The result would be::

        {
            "test.module.ModuleAlias": {
                "id": "test.module.ModuleAlias",
                "name": "Module1",
                "alias": "ModuleAlias"
            },
            "test.module.Module2": {
                "id": "test.module.ModuleAlias",
                "name": "Module2",
                "alias": None
            }
        }

    """
    environments = []

    for element in expression.split(","):
        _element = element.strip()
        if len(_element) == 0:
            continue

        match = _BINDING_ELEMENT_PATTERN.match(_element)
        if match is None:
            continue

        name = match.group("name")
        alias = match.group("alias")

        # Determine element ID if an alias is set
        element_id = alias if alias is not None else name
        if element_id == "*":
            wildcards_number += 1
            element_id = "WILDCARD_{0}".format(wildcards_number)

        element_id = "{module}.{id}".format(
            module=module_id, id=element_id
        )

        _module = {
            "id": element_id,
            "name": name,
            "alias": alias
        }
        environments.append(_module)

    return environments, wildcards_number



# :coding: utf-8

import collections

from sphinx.directives import Directive
from sphinx import addnodes
import docutils.parsers.rst.directives
import docutils.nodes

# import rst_generator
from .rst_generator import (
    get_rst_class_elements,
    get_rst_function_elements,
    get_rst_data_elements,
    get_rst_export_elements,
    rst_string
)


def _parse_members(argument):
    """Convert the :members: options to module directive."""
    if argument is None:
        return True
    return [arg.strip() for arg in argument.split(",")]


class AutoModuleDirective(Directive):
    """Directive to render :term:`Javascript` module documentation.

    The unique argument should be the identifier of the module element.

    .. sourcecode:: rest

        .. js:automodule:: module.test

    The available options are:

    * members:
        This option can be boolean if no arguments are given to indicate that
        all members should be documented, or a white list of member names to
        display.

    * skip-description:
        Indicate whether the module description should be skipped.

    * skip-data-value:
        Indicate whether data values within the module should be skipped.

    * skip-attribute-value:
        Indicate whether attribute values within the module should be skipped.

    * undoc-members:
        Indicate whether members with no docstrings should be displayed.

    * private-members:
        Indicate whether private members (with a name starting with an
        underscore) should be displayed.

    * module-alias:
        String element to replace the module name.

    * module-path-alias:
        String element to replace the module path.

    * force-partial-import:
        Indicate whether each import statement display within the module
        should be indicated with partial import.

    .. seealso::

        :ref:`directive/automodule`

    .. seealso::

        :ref:`configuration/js_module_options`


    """
    #: Only the object id argument is required
    required_arguments = 1

    #: No optional argument is available
    optional_arguments = 0

    #: Content is automatically generated and can not be manually entered
    has_content = False

    #: Nested element is automatically generated and can not be manually entered
    allow_nesting = False

    #: module options
    option_spec = {
        "members": _parse_members,
        "skip-description": lambda x: True,
        "skip-data-value": lambda x: True,
        "skip-attribute-value": lambda x: True,
        "undoc-members": lambda x: True,
        "private-members": lambda x: True,
        "module-alias": docutils.parsers.rst.directives.unchanged_required,
        "module-path-alias": docutils.parsers.rst.directives.unchanged_required,
        "force-partial-import": lambda x: True,
    }

    def run(self):
        """Run the directive."""
        # The signature is always the first argument.
        signature = self.arguments[0]

        js_env = self.state.document.settings.env.app.config.js_environment
        if signature not in js_env["module"].keys():
            raise self.error(
                "The module id is unavailable: {signature}".format(
                    signature=signature
                )
            )

        env = self.state.document.settings.env
        module_environment = js_env["module"][signature]

        # Automatic boolean options
        options = self.state.document.settings.env.config.js_module_options

        # Update references
        ref_context = self.state.document.settings.env.ref_context
        ref_context["js:module"] = module_environment["id"]
        ref_context["js:object"] = "module"

        nodes = []

        # Add target to reference this module
        module_id = module_environment["id"]
        env.domaindata["js"]["modules"][module_id] = env.docname
        env.domaindata['js']["objects"][module_id] = (env.docname, "module")
        target_node = docutils.nodes.target(
            "", "", ids=["module-" + module_id], ismod=True
        )

        self.state.document.note_explicit_target(target_node)
        nodes.append(target_node)
        index_text = "{0} (module)".format(module_id)
        index_node = addnodes.index(
            entries=[("single", index_text, "module-" + module_id, "", None)]
        )
        nodes.append(index_node)

        file_environment = self._file_environment(module_environment)
        description = file_environment["description"]

        skip_description = self.options.get(
            "skip-description", "skip-description" in options
        )

        if description and not skip_description:
            rst_element = rst_string(description)
            node = docutils.nodes.paragraph()
            self.state.nested_parse(rst_element, 0, node)
            nodes.append(node)

        members = self.options.get("members", "members" in options)
        if members:
            whitelist = (
                members if isinstance(members, collections.Iterable) else None
            )

            nodes += self.generate_members(
                module_environment, options, whitelist_names=whitelist
            )

        return nodes

    def _file_environment(self, module_environment):
        """Get the file environment from the *module_environment*.
        """
        js_env = self.state.document.settings.env.app.config.js_environment

        file_id = module_environment["file_id"]
        file_environment = js_env["file"][file_id]
        return file_environment

    def generate_members(
        self, module_environment, options, whitelist_names=None
    ):
        """Generate a list of member nodes from *module_environment*

        *options* is the dictionary of module options that can affect the
        display of members

        *whitelist_names* is an optional list of element names that
        should be displayed exclusively.

        """
        js_env = self.state.document.settings.env.app.config.js_environment
        file_environment = self._file_environment(module_environment)

        module_name = self.options.get(
            "module-alias", module_environment["name"]
        )

        module_path_name = self.options.get(
            "module-path-alias", module_environment["path"]
        )

        # Options manually set
        undoc_members = self.options.get(
            "undoc-members", "undoc-members" in options
        )
        private_members = self.options.get(
            "private-members", "private-members" in options
        )

        rst_elements = {}

        # Gather classes
        rst_elements = get_rst_class_elements(
            file_environment, module_name, module_path_name,
            whitelist_names=whitelist_names,
            undocumented_members=undoc_members,
            private_members=private_members,
            force_partial_import=self.options.get(
                "force-partial-import", False
            ),
            skip_attribute_value=self.options.get(
                "skip-attribute-value", "skip-attribute-value" in options
            ),
            rst_elements=rst_elements
        )

        # Gather functions
        rst_elements = get_rst_function_elements(
            file_environment, module_name, module_path_name,
            whitelist_names=whitelist_names,
            undocumented_members=undoc_members,
            private_members=private_members,
            force_partial_import=self.options.get(
                "force-partial-import", False
            ),
            rst_elements=rst_elements,
        )

        # Gather variables
        rst_elements = get_rst_data_elements(
            file_environment, module_name, module_path_name,
            whitelist_names=whitelist_names,
            blacklist_ids=file_environment["function"].keys(),
            undocumented_members=undoc_members,
            private_members=private_members,
            force_partial_import=self.options.get(
                "force-partial-import", False
            ),
            skip_value=self.options.get(
                "skip-data-value", "skip-data-value" in options
            ),
            rst_elements=rst_elements,
        )

        # Gather exported elements
        rst_elements = get_rst_export_elements(
            file_environment, js_env, module_name, module_path_name,
            skip_data_value=self.options.get(
                "skip-data-value", "skip-data-value" in options
            ),
            skip_attribute_value=self.options.get(
                "skip-attribute-value", "skip-attribute-value" in options
            ),
            rst_elements=rst_elements,
        )

        # Add content while respecting the line order
        nodes = []

        for line_number in sorted(rst_elements.keys()):
            for rst_element in rst_elements[line_number]:
                node = docutils.nodes.paragraph()
                self.state.nested_parse(rst_element, 0, node)

                nodes.append(node)

        return nodes



# :coding: utf-8

import os

from ._version import __version__

from .directive.js_data import AutoDataDirective
from .directive.js_function import AutoFunctionDirective
from .directive.js_class import (
    AutoClassDirective, AutoMethodDirective, AutoAttributeDirective
)
from .directive.js_module import AutoModuleDirective

from .viewcode import ViewCode
from .parser import fetch_environment


def setup(app):
    """Register callbacks and directives."""
    app.add_config_value("js_source", None, True)
    app.add_config_value("js_sources", [], True)
    app.add_config_value("js_environment", None, True)
    app.add_config_value("js_class_options", [], True)
    app.add_config_value("js_module_options", [], True)

    app.connect("builder-inited", fetch_javascript_environment)
    app.connect("doctree-read", ViewCode.add_source_code_links)
    app.connect("html-collect-pages", ViewCode.create_code_pages)
    app.connect("missing-reference", ViewCode.create_missing_code_link)

    app.add_directive_to_domain("js", "autodata", AutoDataDirective)
    app.add_directive_to_domain("js", "autofunction", AutoFunctionDirective)
    app.add_directive_to_domain("js", "autoclass", AutoClassDirective)
    app.add_directive_to_domain("js", "automethod", AutoMethodDirective)
    app.add_directive_to_domain("js", "autoattribute", AutoAttributeDirective)
    app.add_directive_to_domain("js", "automodule", AutoModuleDirective)

    return {
        "version": __version__
    }


def fetch_javascript_environment(app):
    """Fetch the :term:`Javascript` environment from the *app* configuration.

    If the **js_environment** configuration is not provided, attempt to parse
    the path provided via the **js_source** or **js_sources** configuration
    value.

    This function is called with the ``builder-inited`` Sphinx event, emitted
    when the builder object is created.

    .. seealso::

        :ref:`configuration`

    """
    if app.config.js_environment is not None:
        return

    if app.config.js_source is not None:
        path = os.path.abspath(app.config.js_source)
        app.config.js_environment = fetch_environment(path)

    elif len(app.config.js_sources) > 0:
        app.config.js_environment = {}

        for path in app.config.js_sources:
            path = os.path.abspath(path)

            _environment = fetch_environment(path)
            for key in _environment.keys():
                app.config.js_environment.setdefault(key, {})
                app.config.js_environment[key].update(_environment[key])

    else:
        raise RuntimeError(
            "Either the 'js_source' or the 'js_sources' configuration value "
            "must be provided."
        )



# :coding: utf-8

import os
import shutil
import tempfile

import pytest


@pytest.fixture()
def temporary_directory(request):
    """Return a temporary directory path."""
    path = tempfile.mkdtemp()

    def cleanup():
        """Remove temporary directory."""
        shutil.rmtree(path)

    request.addfinalizer(cleanup)

    return path


@pytest.fixture()
def doc_folder(temporary_directory):
    path = os.path.join(temporary_directory, "doc")
    os.makedirs(path)

    # Create the js code source folder
    js_source = os.path.join(temporary_directory, "doc", "example")
    os.makedirs(js_source)

    # Create minimal conf.py file
    conf_file = os.path.join(path, "conf.py")

    with open(conf_file, "w") as f:
        f.write(
            "# :coding: utf-8\n"
            "extensions=['champollion']\n"
            "source_suffix = '.rst'\n"
            "master_doc = 'index'\n"
            "author = u'Jeremy Retailleau'\n"
            "exclude_patterns = ['Thumbs.db', '.DS_Store']\n"
            "js_source='{}/example'".format(path)
        )

    return path


@pytest.fixture()
def doc_folder_with_class_members(temporary_directory):
    path = os.path.join(temporary_directory, "doc")
    os.makedirs(path)

    # Create the js code source folder
    js_source = os.path.join(temporary_directory, "doc", "example")
    os.makedirs(js_source)

    # Create minimal conf.py file
    conf_file = os.path.join(path, "conf.py")

    with open(conf_file, "w") as f:
        f.write(
            "# :coding: utf-8\n"
            "extensions=['champollion']\n"
            "source_suffix = '.rst'\n"
            "master_doc = 'index'\n"
            "author = u'Jeremy Retailleau'\n"
            "exclude_patterns = ['Thumbs.db', '.DS_Store']\n"
            "js_class_options = ['members']\n"
            "js_source='{}/example'".format(path)
        )

    return path



# :coding: utf-8

import pytest
import champollion.parser.js_function


@pytest.mark.parametrize(
    ("content", "expected"),
    [
        (
            (
                "/** test function */\n"
                "export function doSomething() {\n"
                "    console.log('something');\n"
                "}\n"
            ),
            {
                "test.module.doSomething": {
                    "id": "test.module.doSomething",
                    "module_id": "test.module",
                    "exported": True,
                    "default": False,
                    "name": "doSomething",
                    "anonymous": False,
                    "generator": False,
                    "arguments": [],
                    "line_number": 2,
                    "description": "test function"
                }
            }
        ),
        (
            (
                "/**\n"
                " * test function with one argument.\n"
                " */\n"
                "const doSomething = arg1 => {\n"
                "    console.log('something');\n"
                "};\n"
            ),
            {
                "test.module.doSomething": {
                    "id": "test.module.doSomething",
                    "module_id": "test.module",
                    "exported": False,
                    "default": False,
                    "name": "doSomething",
                    "anonymous": False,
                    "generator": False,
                    "arguments": ["arg1"],
                    "line_number": 4,
                    "description": "test function with one argument."
                }
            }
        ),
        (
            (
                "/**\n"
                " * test function with arguments.\n"
                " */\n"
                "const doSomething = (arg1, arg2) => {\n"
                "    console.log('something');\n"
                "};\n"
            ),
            {
                "test.module.doSomething": {
                    "id": "test.module.doSomething",
                    "module_id": "test.module",
                    "exported": False,
                    "default": False,
                    "name": "doSomething",
                    "anonymous": False,
                    "generator": False,
                    "arguments": ["arg1", "arg2"],
                    "line_number": 4,
                    "description": "test function with arguments."
                }
            }
        ),
        (
            (
                "export const doSomethingWithManyArguments = (\n"
                "  arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9\n"
                ") => {};\n"
            ),
            {
                "test.module.doSomethingWithManyArguments": {
                    "id": "test.module.doSomethingWithManyArguments",
                    "module_id": "test.module",
                    "exported": True,
                    "default": False,
                    "name": "doSomethingWithManyArguments",
                    "anonymous": False,
                    "generator": False,
                    "arguments": [
                        "arg1", "arg2", "arg3", "arg4",
                        "arg5", "arg6", "arg7", "arg8",
                        "arg9"
                    ],
                    "line_number": 1,
                    "description": None
                }
            }
        ),
        (
            (
                "/**\n"
                " * test anonymous function.\n"
                " */\n"
                "export default function (arg1) {\n"
                "    console.log('anonymous function');\n"
                "};\n"
            ),
            {
                "test.module.__ANONYMOUS_FUNCTION__": {
                    "id": "test.module.__ANONYMOUS_FUNCTION__",
                    "module_id": "test.module",
                    "exported": True,
                    "default": True,
                    "name": "__ANONYMOUS_FUNCTION__",
                    "anonymous": True,
                    "generator": False,
                    "arguments": ["arg1"],
                    "line_number": 4,
                    "description": "test anonymous function."
                }
            }
        ),
        (
            (
                "/**\n"
                " * test generator function.\n"
                " */\n"
                "function* generate(arg1, arg2) {\n"
                "    yield something();\n"
                "};\n"
            ),
            {
                "test.module.generate": {
                    "id": "test.module.generate",
                    "module_id": "test.module",
                    "exported": False,
                    "default": False,
                    "name": "generate",
                    "anonymous": False,
                    "generator": True,
                    "arguments": ["arg1", "arg2"],
                    "line_number": 4,
                    "description": "test generator function."
                },
            }
        ),
        (
            (
                "const doSomethingElse= function doSomething(arg) {\n"
                "    console.log('something');\n"
                "}\n"
            ),
            {
                "test.module.doSomethingElse": {
                    "id": "test.module.doSomethingElse",
                    "module_id": "test.module",
                    "exported": False,
                    "default": False,
                    "name": "doSomethingElse",
                    "anonymous": False,
                    "generator": False,
                    "arguments": ["arg"],
                    "line_number": 1,
                    "description": None
                }
            }
        ),
        (
            (
                "export const doSomethingElse= function* doSomething() {\n"
                "    console.log('something');\n"
                "}\n"
            ),
            {
                "test.module.doSomethingElse": {
                    "id": "test.module.doSomethingElse",
                    "module_id": "test.module",
                    "exported": True,
                    "default": False,
                    "name": "doSomethingElse",
                    "anonymous": False,
                    "generator": True,
                    "arguments": [],
                    "line_number": 1,
                    "description": None
                }
            }
        )
    ],
    ids=[
        "valid exported function",
        "valid arrow-type function one argument",
        "valid arrow-type function with two arguments",
        "valid arrow-type function with several arguments",
        "valid anonymous function",
        "valid generator function",
        "valid function expression",
        "valid generator function expression"
    ]
)
def test_get_function_environment(content, expected):
    """Return function environment from content."""
    assert champollion.parser.js_function.fetch_environment(
        content, "test.module"
    ) == expected


@pytest.mark.parametrize(
    ("content", "expected"),
    [
        (
            "export default function(arg1) {}",
            {
                "arguments": "arg1",
                "data_name": None,
                "default": "default ",
                "export": "export ",
                "function_name": None,
                "generator": None,
                "start_regex": ""
            }
        ),
        (
            "export function doSomething_1() {}",
            {
                "arguments": "",
                "data_name": None,
                "default": None,
                "export": "export ",
                "function_name": "doSomething_1",
                "generator": None,
                "start_regex": ""
            }
        ),
        (
            (
                "function doSomething_Else(\n"
                "    arg1, arg2, arg3, arg3, arg4, arg5,\n"
                "    arg6\n"
                ") {\n"
                "    console.log('test')\n"
                "}"
            ),
            {
                "arguments": "arg1, arg2, arg3, arg3, arg4, arg5,\n    arg6",
                "data_name": None,
                "default": None,
                "export": None,
                "function_name": "doSomething_Else",
                "generator": None,
                "start_regex": ""
            }
        ),
        (
            (
                "const aFunction=function saySomething(text) {\n"
                "    console.log(text)\n"
                "}"
            ),
            {
                "arguments": "text",
                "data_name": "aFunction",
                "default": None,
                "export": None,
                "function_name": "saySomething",
                "generator": None,
                "start_regex": ""
            }
        ),
        (
            (
                "let aFunction=function* saySomething(text) {\n"
                "    console.log(text)\n"
                "}"
            ),
            {
                "arguments": "text",
                "data_name": "aFunction",
                "default": None,
                "export": None,
                "function_name": "saySomething",
                "generator": "* ",
                "start_regex": ""
            }
        ),
        (
            (
                "function* saySomething(text) {\n"
                "    console.log(text)\n"
                "}"
            ),
            {
                "arguments": "text",
                "data_name": None,
                "default": None,
                "export": None,
                "function_name": "saySomething",
                "generator": "* ",
                "start_regex": ""
            }
        ),
        (
            "const test = 'const test = function() {}'",
            None
        )
    ],
    ids=[
        "valid anonymous function",
        "valid named function",
        "valid named function with multiple arguments",
        "valid function expression",
        "valid function generator expression",
        "valid function generator",
        "valid function string",
    ]
)
def test_function_pattern(content, expected):
    """Match a function."""
    match = champollion.parser.js_function._FUNCTION_PATTERN.search(content)
    if expected is None:
        assert match is None
    else:
        assert match.groupdict() == expected


@pytest.mark.parametrize(
    ("content", "expected"),
    [
        (
            "const arrow_type_function = (arg1) => {};",
            {
                "arguments": "arg1",
                "single_argument": None,
                "function_name": "arrow_type_function",
                "default": None,
                "export": None,
                "start_regex": ""
            }
        ),
        (
            "let arrow_type_function2 = arg1 => {};",
            {
                "arguments": None,
                "single_argument": "arg1",
                "function_name": "arrow_type_function2",
                "default": None,
                "export": None,
                "start_regex": ""
            }
        ),
        (
            "export const arrow_type_function = (arg1) => {};",
            {
                "arguments": "arg1",
                "single_argument": None,
                "function_name": "arrow_type_function",
                "default": None,
                "export": "export ",
                "start_regex": ""
            }
        ),
        (
            "export default var arrow_type_function3 = (arg1, arg2) => {};",
            {
                "arguments": "arg1, arg2",
                "single_argument": None,
                "function_name": "arrow_type_function3",
                "default": "default ",
                "export": "export ",
                "start_regex": ""
            }
        ),
        (
            (
                "export const arrow_type_function = (\n"
                "    arg1, arg2, arg3, arg4, arg5, agr6,\n"
                "    arg7, arg8\n"
                ") => {\n"
                "    console.log('youpi');\n"
                "};\n"
            ),
            {
                "arguments": (
                    "arg1, arg2, arg3, arg4, arg5, agr6,\n    arg7, arg8"
                ),
                "single_argument": None,
                "function_name": "arrow_type_function",
                "default": None,
                "export": "export ",
                "start_regex": ""
            }
        ),
        (
            "arrow_type_function = (arg1) => {}",
            None
        ),
        (
            "const test = 'const arrow_type_function = (arg1) => {}'",
            None
        ),
        (
            "(arg1) => {}",
            None
        ),
        (
            "const arrow_type_function = arg1, arg2 => {};",
            None
        ),
    ],
    ids=[
        "valid function with one argument and brackets",
        "valid function with one argument and no brackets",
        "valid exported function",
        "valid function with two arguments",
        "valid function with multiple arguments",
        "invalid arrow-type function without type",
        "invalid arrow-type function string",
        "invalid unassigned arrow-type function",
        "invalid arrow-type function with multiple argument and no brackets",
    ]
)
def test_function_arrow_pattern(content, expected):
    """Match an arrow-type function."""
    match = champollion.parser.js_function._FUNCTION_ARROW_PATTERN.search(
        content
    )
    if expected is None:
        assert match is None
    else:
        assert match.groupdict() == expected


@pytest.mark.parametrize(
    ("content", "expected"),
    [
        (
            "connect(arg1, arg2);",
            {
                "arguments": "arg1, arg2",
                "default": None,
                "export": None,
                "function_name": "connect",
                "start_regex": ""
            }
        ),
        (
            "connect("
            "    arg1, arg2, arg3, arg4, arg5, arg6,\n"
            "    arg7,\n"
            ");",
            {
                "arguments": "arg1, arg2, arg3, arg4, arg5, arg6,\n    arg7,",
                "default": None,
                "export": None,
                "function_name": "connect",
                "start_regex": ""
            }
        ),
        (
            "export   connect(arg1, arg2);",
            {
                "arguments": "arg1, arg2",
                "default": None,
                "export": "export   ",
                "function_name": "connect",
                "start_regex": ""
            }
        ),
        (
            "\n"
            "export default connect(arg)()",
            {
                "arguments": "arg",
                "default": "default ",
                "export": "export ",
                "function_name": "connect",
                "start_regex": "\n"
            }
        )
    ],
    ids=[
        "valid function",
        "valid function with multiple arguments",
        "valid exported function",
        "valid exported default function",
    ]
)
def test_imported_function_pattern(content, expected):
    """Match an imported function."""
    match = champollion.parser.js_function._IMPORTED_FUNCTION_PATTERN.search(
        content
    )
    if expected is None:
        assert match is None
    else:
        assert match.groupdict() == expected
